"""
Modules for calculating theoretical solar-cell efficiency.

Please find more detailsin:
https://pubs.acs.org/doi/abs/10.1021/acs.chemmater.9b02166
"""
import numpy as np
import os
from scipy.interpolate import interp1d
from numpy import interp
import scipy.constants as constants
from scipy.integrate import simps
import matplotlib.pyplot as plt


class SolarEfficiency(object):
    """Calculate theoretical solar-efficiency using SLME or SQ approach."""

    def __init__(self, formalism="slme"):
        """Use SLME or SQ formalisms."""
        self.formalism = formalism

    def calculate_SQ(
        self,
        bandgap_ev,
        temperature=300,
        fr=1,
        plot_current_voltage=False,
        filename="sq.png",
    ):
        """
        Calcualte efficeincy using shockley queisser formalism.

        Requires only two inputs unlike SLME.
        Args:
              bandgap_ev: bandga in electron-volt

              temperature: temperature in K
        """
        # Defining constants for tidy equations
        c = constants.c  # speed of light, m/s
        h = constants.h  # Planck's constant J*s (W)
        h_e = constants.h / constants.e  # Planck's constant eV*s
        k = constants.k  # Boltzmann's constant J/K
        e = constants.e  # Coulomb

        # Load the Air Mass 1.5 Global tilt solar spectrum
        solar_spectrum_data_file = str(
            os.path.join(os.path.dirname(__file__), "am1.5G.dat")
        )
        solar_spectra_wavelength, solar_spectra_irradiance = np.loadtxt(
            solar_spectrum_data_file, usecols=[0, 1], unpack=True, skiprows=2
        )

        solar_spectra_wavelength_meters = solar_spectra_wavelength * 1e-9

        # need to convert solar irradiance from Power/m**2(nm) into
        # photon#/s*m**2(nm) power is Watt, which is Joule / s
        # E = hc/wavelength
        # at each wavelength, Power * (wavelength(m)/(h(Js)*c(m/s))) = ph#/s
        solar_spectra_photon_flux = solar_spectra_irradiance * (
            solar_spectra_wavelength_meters / (h * c)
        )

        # Calculation of total solar power incoming
        power_in = simps(solar_spectra_irradiance, solar_spectra_wavelength)

        # calculation of blackbody irradiance spectra
        # units of W/(m**3), different than solar_spectra_irradiance!!! (This
        # is intentional, it is for convenience)
        blackbody_irradiance = (
            2.0 * h * c ** 2 / (solar_spectra_wavelength_meters ** 5)
        ) * (
            1.0
            / (
                (
                    np.exp(
                        h
                        * c
                        / (solar_spectra_wavelength_meters * k * temperature)
                    )
                )
                - 1.0
            )
        )

        # I've removed a pi in the equation above - Marnik Bercx

        # Convert the irradiance from Power/m**2(m) into photon#/s*m**2(m)
        blackbody_photon_flux = blackbody_irradiance * (
            solar_spectra_wavelength_meters / (h * c)
        )

        # absorbance interpolation onto each solar spectrum wavelength

        # Get the bandgap in wavelength in meters
        bandgap_wavelength = h_e * c / bandgap_ev

        # Only take the part of the wavelength-dependent solar spectrum and
        # blackbody spectrum below the bandgap wavelength
        bandgap_index = np.searchsorted(
            solar_spectra_wavelength_meters, bandgap_wavelength
        )

        bandgap_irradiance = interp(
            np.array([bandgap_wavelength]),
            solar_spectra_wavelength_meters,
            solar_spectra_photon_flux,
        )

        bandgap_blackbody = (
            (2.0 * h * c ** 2 / (bandgap_wavelength ** 5))
            * (
                1.0
                / (
                    (np.exp(h * c / (bandgap_wavelength * k * temperature)))
                    - 1.0
                )
            )
            * (bandgap_wavelength / (h * c))
        )

        integration_wavelength = np.concatenate(
            (
                solar_spectra_wavelength_meters[:bandgap_index],
                np.array([bandgap_wavelength]),
            ),
            axis=0,
        )

        integration_solar_flux = np.concatenate(
            (solar_spectra_photon_flux[:bandgap_index], bandgap_irradiance),
            axis=0,
        )

        integration_blackbody = np.concatenate(
            (
                blackbody_photon_flux[:bandgap_index],
                np.array([bandgap_blackbody]),
            ),
            axis=0,
        )

        #  Numerically integrating irradiance over wavelength array
        # Note: elementary charge, not math e!  ## units of A/m**2   W/(V*m**2)
        J_0_r = (
            e * np.pi * simps(integration_blackbody, integration_wavelength)
        )

        J_0 = J_0_r / fr

        #  Numerically integrating irradiance over wavelength array
        # elementary charge, not math e!  ### units of A/m**2   W/(V*m**2)
        J_sc = e * simps(integration_solar_flux * 1e9, integration_wavelength)

        #    J[i] = J_sc - J_0*(1 - exp( e*V[i]/(k*T) ) )
        #   #This formula from the original paper has a typo!!
        #    J[i] = J_sc - J_0*(exp( e*V[i]/(k*T) ) - 1)
        #   #Bercx chapter and papers have the correct formula (well,
        #   the correction on one paper)
        def J(V):
            J = J_sc - J_0 * (np.exp(e * V / (k * temperature)) - 1.0)
            return J

        def power(V):
            p = J(V) * V
            return p

        # A more primitive, but perfectly robust way of getting a reasonable
        # estimate for the maximum power.
        test_voltage = 0
        voltage_step = 0.001
        while power(test_voltage + voltage_step) > power(test_voltage):
            test_voltage += voltage_step

        max_power = power(test_voltage)

        # Calculate the maximized efficience
        efficiency = max_power / power_in

        # This segment isn't needed for functionality at all, but can display a
        # plot showing how the maximization of power by choosing the optimal
        # voltage value works
        if plot_current_voltage:
            V = np.linspace(0, 2, 200)
            plt.plot(V, J(V))
            plt.plot(V, power(V), linestyle="--")
            plt.savefig(filename)
            print(max_power)

        return efficiency

    def slme(
        self,
        material_energy_for_absorbance_data,
        material_absorbance_data,
        material_direct_allowed_gap,
        material_indirect_gap,
        thickness=50e-6,
        temperature=293.15,
        absorbance_in_inverse_centimeters=False,
        cut_off_absorbance_below_direct_allowed_gap=True,
        plot_current_voltage=False,
        filename="slme.png",
    ):
        """
        Calculate spectroscopic limited maximum efficiency.

        Reuires more info than SQ.
        Args:
            material_energy_for_absorbance_data:
            energy grid for absorbance data

            material_absorbance_data: absorption coefficient in m^-1

            material_direct_allowed_gap: direct bandgap in eV

            material_indirect_gap: indirect bandgap in eV

            thickness: thickness of the material in m

            temperature: working temperature in K

            absorbance_in_inverse_centimeters:
            whether the absorbance data is in the unit of cm^-1

            cut_off_absorbance_below_direct_allowed_gap:
            whether to discard all absorption below bandgap

            plot_current_voltage: whether to plot the current-voltage curve

        Returns:
            The calculated maximum efficiency.
        """
        # Defining constants for tidy equations
        c = constants.c  # speed of light, m/s
        h = constants.h  # Planck's constant J*s (W)
        h_e = constants.h / constants.e  # Planck's constant eV*s
        k = constants.k  # Boltzmann's constant J/K
        k_e = constants.k / constants.e  # Boltzmann's constant eV/K
        e = constants.e  # Coulomb

        # Make sure the absorption coefficient has the right units (m^{-1})
        if absorbance_in_inverse_centimeters:
            material_absorbance_data = material_absorbance_data * 100

        # Load the Air Mass 1.5 Global tilt solar spectrum
        solar_spectrum_data_file = str(
            os.path.join(os.path.dirname(__file__), "am1.5G.dat")
        )

        solar_spectra_wavelength, solar_spectra_irradiance = np.loadtxt(
            solar_spectrum_data_file, usecols=[0, 1], unpack=True, skiprows=2
        )

        solar_spectra_wavelength_meters = solar_spectra_wavelength * 1e-9

        delta = material_direct_allowed_gap - material_indirect_gap
        fr = np.exp(-delta / (k_e * temperature))

        # need to convert solar irradiance from Power/m**2(nm) into
        # photon#/s*m**2(nm) power is Watt, which is Joule / s
        # E = hc/wavelength
        # at each wavelength, Power * (wavelength(m)/(h(Js)*c(m/s))) = ph#/s
        solar_spectra_photon_flux = solar_spectra_irradiance * (
            solar_spectra_wavelength_meters / (h * c)
        )

        # Calculation of total solar power incoming
        power_in = simps(solar_spectra_irradiance, solar_spectra_wavelength)

        # calculation of blackbody irradiance spectra
        # units of W/(m**3), different than solar_spectra_irradiance!!! (This
        # is intentional, it is for convenience)
        blackbody_irradiance = (
            2.0 * h * c ** 2 / (solar_spectra_wavelength_meters ** 5)
        ) * (
            1.0
            / (
                (
                    np.exp(
                        h
                        * c
                        / (solar_spectra_wavelength_meters * k * temperature)
                    )
                )
                - 1.0
            )
        )

        # I've removed a pi in the equation above - Marnik Bercx

        # Convert the irradiance from Power/m**2(m) into photon#/s*m**2(m)
        blackbody_photon_flux = blackbody_irradiance * (
            solar_spectra_wavelength_meters / (h * c)
        )

        # units of nm
        material_wavelength_for_absorbance_data = (
            (c * h_e) / (material_energy_for_absorbance_data + 0.00000001)
        ) * 10 ** 9

        # absorbance interpolation onto each solar spectrum wavelength

        # creates cubic spline interpolating function, set up to use end values
        #  as the guesses if leaving the region where data exists
        material_absorbance = interp1d(
            material_wavelength_for_absorbance_data,
            material_absorbance_data,
            kind="cubic",
            fill_value=(
                material_absorbance_data[0],
                material_absorbance_data[-1],
            ),
            bounds_error=False,
        )

        material_interpolated_absorbance = np.zeros(
            len(solar_spectra_wavelength_meters)
        )
        for i in range(0, len(solar_spectra_wavelength_meters)):
            # Cutting off absorption data below the gap. This is done to deal
            # with VASPs broadening of the calculated absorption data

            if (
                solar_spectra_wavelength[i]
                < 1e9 * ((c * h_e) / material_direct_allowed_gap)
                or not cut_off_absorbance_below_direct_allowed_gap
            ):
                material_interpolated_absorbance[i] = material_absorbance(
                    solar_spectra_wavelength[i]
                )

        absorbed_by_wavelength = 1.0 - np.exp(
            -2.0 * material_interpolated_absorbance * thickness
        )

        #  Numerically integrating irradiance over wavelength array
        # Note: elementary charge, not math e!  ## units of A/m**2   W/(V*m**2)
        J_0_r = (
            e
            * np.pi
            * simps(
                blackbody_photon_flux * absorbed_by_wavelength,
                solar_spectra_wavelength_meters,
            )
        )

        J_0 = J_0_r / fr

        #  Numerically integrating irradiance over wavelength array
        # elementary charge, not math e!  ### units of A/m**2   W/(V*m**2)
        J_sc = e * simps(
            solar_spectra_photon_flux * absorbed_by_wavelength,
            solar_spectra_wavelength,
        )

        #    J[i] = J_sc - J_0*(1 - exp( e*V[i]/(k*T) ) )
        #   #This formula from the original paper has a typo!!
        #    J[i] = J_sc - J_0*(exp( e*V[i]/(k*T) ) - 1)
        #   #Bercx chapter and papers have the correct formula (well,
        #   the correction on one paper)
        def J(V):
            J = J_sc - J_0 * (np.exp(e * V / (k * temperature)) - 1.0)
            return J

        def power(V):
            p = J(V) * V
            return p

        # A more primitive, but perfectly robust way of getting a reasonable
        # estimate for the maximum power.
        test_voltage = 0
        voltage_step = 0.001
        while power(test_voltage + voltage_step) > power(test_voltage):
            test_voltage += voltage_step

        max_power = power(test_voltage)

        # Calculate the maximized efficience
        efficiency = max_power / power_in

        # This segment isn't needed for functionality at all, but can display a
        # plot showing how the maximization of power by choosing the optimal
        # voltage value works
        if plot_current_voltage:
            V = np.linspace(0, 2, 200)
            plt.plot(V, J(V))
            plt.plot(V, power(V), linestyle="--")
            plt.savefig(filename)
            plt.close()
            # print(power(V_Pmax))

        return efficiency


"""
if __name__ == "__main__":
    from jarvis.io.vasp.outputs import Vasprun
    v = Vasprun(
        "/rk2/knc6/JARVIS-DFT/Elements-bulkk/mp-149_bulk_PBEBO/MAIN-MBJ-bulk@mp-149/vasprun.xml"
    )
    dirgap = v.get_dir_gap
    indirgap = v.get_indir_gap
    en, abz = v.avg_absorption_coefficient
    abz = abz * 100
    eff = SolarEfficiency().slme(
        en, abz, indirgap, indirgap, plot_current_voltage=False
    )
    print("SLME", 100 * eff)
    eff = SolarEfficiency().calculate_SQ(indirgap)
    print("SQ", 100 * eff)
"""



"""Modules to generate input files for phonopy using Atoms object."""

import numpy as np
from jarvis.analysis.structure.spacegroup import Spacegroup3D
from jarvis.core.kpoints import Kpoints3D


class PhonopyInputs(object):
    """
    Generate input files for phonopy post-processing.

    Make sure to obtain FORCE_CONSTANTS first
    E.g. run phonopy --fc vasprun.xml first.
    """

    def __init__(self, atoms):
        """Intialize with jarvis.core.Atoms class."""
        self.atoms = atoms
        self.tag = " ".join(list(set(atoms.elements)))

    def prim_axis(self):
        """
        Use when taking conventional cell.

        Such as elastic constant calculations.
        """
        prim = self.atoms.get_primitive_atoms
        if prim.num_atoms < self.atoms.num_atoms:
            spaceg = Spacegroup3D(self.atoms)
            latt = spaceg.lattice_system
            sgp = spaceg.space_group_symbol
            if latt == "rhombohedral":
                transf = (
                    np.array(
                        [[2, -1, -1], [1, 1, -2], [1, 1, 1]], dtype=np.float
                    )
                    / 3
                )
            elif "I" in sgp:
                transf = (
                    np.array(
                        [[-1, 1, 1], [1, -1, 1], [1, 1, -1]], dtype=np.float
                    )
                    / 2
                )

            elif "F" in sgp:
                transf = (
                    np.array([[0, 1, 1], [1, 0, 1], [1, 1, 0]], dtype=np.float)
                    / 2
                )
            elif "A" in sgp:
                transf = (
                    np.array(
                        [[2, 0, 0], [0, 1, -1], [0, 1, 1]], dtype=np.float
                    )
                    / 2
                )
            elif "C" in sgp:
                if latt == "monoclinic":
                    transf = (
                        np.array(
                            [[1, 1, 0], [-1, 1, 0], [0, 0, 2]], dtype=np.float
                        )
                        / 2
                    )
                else:
                    transf = (
                        np.array(
                            [[1, -1, 0], [1, 1, 0], [0, 0, 2]], dtype=np.float
                        )
                        / 2
                    )
        else:
            transf = np.eye(3)
        ax = ""
        for el in transf:
            ax = (
                str(ax)
                + str(" ")
                + str(el[0])
                + str(" ")
                + str(el[1])
                + str(" ")
                + str(el[2])
            )
        ax_line = str("PRIMITIVE_AXIS = ") + str(ax) + "\n"
        return ax_line

    def mesh_dos(
        self,
        filename="meshdos.conf",
        dim=[1, 1, 1],
        factor=521.471,
        grid=[31, 31, 31],
    ):
        """
        Use for DOS and thermal properties.

        phonopy -p meshdos.conf
        phonopy -t meshdos.conf
        """
        mesh = open(filename, "w")
        line = str("FORCE_CONSTANTS = READ") + "\n"
        mesh.write(line)
        line = str("FREQUENCY_CONVERSION_FACTOR = ") + str(factor) + "\n"
        mesh.write(line)
        line = str("ATOM_NAME = ") + str(self.tag) + "\n"
        mesh.write(line)
        line = str("DIM = ") + " ".join(map(str, dim)) + "\n"
        mesh.write(line)
        line = str("MP = ") + " ".join(map(str, grid)) + "\n"
        mesh.write(line)
        mesh.close()

    def mesh_bands(
        self,
        dim=[1, 1, 1],
        filename="band.conf",
        factor=521.471,
        line_density=20,
    ):
        """
        Use for making phonon bandstructure plot.

        After running phonopy -p bandd.conf, bandplot -o PBAND.png
        """
        kpoints = Kpoints3D().kpath(self.atoms, line_density=line_density)
        all_kp = kpoints._kpoints
        labels = kpoints._labels
        all_labels = ""
        all_lines = ""
        for lb in labels:
            if lb == "":
                lb = None
            all_labels = all_labels + str(lb) + str(" ")
        for k in all_kp:
            all_lines = (
                all_lines
                + str(k[0])
                + str(" ")
                + str(k[1])
                + str(" ")
                + str(k[2])
                + str(" ")
            )
        file = open(filename, "w")
        line = str("FREQUENCY_CONVERSION_FACTOR = ") + str(factor) + "\n"
        file.write(line)
        ax_line = self.prim_axis()
        file.write(ax_line)
        line = str("ATOM_NAME = ") + str(self.tag) + "\n"
        file.write(line)
        line = str("DIM = ") + " ".join(map(str, dim)) + "\n"
        file.write(line)
        line = str("FORCE_CONSTANTS = READ") + "\n"
        file.write(line)
        line = str("BAND= ") + str(all_lines) + "\n"
        file.write(line)
        line = str("BAND_LABELS= ") + str(all_labels) + "\n"
        file.write(line)
        file.close()

    def mesh_irreps(
        self, factor=521.471, filename="irreps.conf", dim=[1, 1, 1], tol=1e-3
    ):
        """Use for irreducible phonon representation."""
        file = open(filename, "w")
        line = str("IRREPS = 0 0 0 ") + str(tol) + "\n"
        file.write(line)
        line = str("FREQUENCY_CONVERSION_FACTOR = ") + str(factor) + "\n"
        file.write(line)
        ax_line = self.prim_axis()
        file.write(ax_line)
        line = str("SHOW_IRREPS = .TRUE.") + "\n"
        file.write(line)
        line = str("DIM = ") + " ".join(map(str, dim)) + "\n"
        file.write(line)
        line = str("FORCE_CONSTANTS = READ") + "\n"
        file.write(line)
        file.close()

    def animation(self, factor=521.471, filename="anim.conf", dim=[1, 1, 1]):
        """Use for atom dancing/movement animation."""
        anim = open(filename, "w")
        line = str("FREQUENCY_CONVERSION_FACTOR = ") + str(factor) + "\n"
        anim.write(line)
        line = str("FORCE_CONSTANTS = READ") + "\n"
        anim.write(line)
        line = str("ATOM_NAME = ") + str(self.tag) + "\n"
        anim.write(line)
        line = str("ANIME_TYPE = JMOL") + "\n"
        anim.write(line)
        line = str("ANIME = 0 7 0  0 0 0") + "\n"
        anim.write(line)
        line = str("DIM = ") + " ".join(map(str, dim)) + "\n"
        anim.write(line)
        anim.close()

    def generate_all_files(self):
        """Generate all the input files mentioned above."""
        self.mesh_dos()
        self.mesh_bands()
        self.mesh_irreps()
        self.animation()


"""
if __name__ == "__main__":
    from jarvis.core.atoms import Atoms
    box = [[2.715, 2.715, 0], [0, 2.715, 2.715], [2.715, 0, 2.715]]
    coords = [[0, 0, 0], [0.25, 0.25, 0.25]]
    elements = ["Si", "Si"]
    Si = Atoms(lattice_mat=box, coords=coords, elements=elements)
    PhonopyInputs(atoms=Si).generate_all_files()
"""



from jarvis.ai.uncertainty import lgbm_quantile_uncertainty as uq
from jarvis.ai.pkgs.utils import get_ml_data
import os

# STEP-1: Getting Data
# ********************
property = "exfoliation_energy"

x, y, jid = get_ml_data(dataset="cfid_3d", ml_property=property)

x = x[0:100]
y = y[0:100]
jid = jid[0:100]

# STEP-2: Quantile regression to make predictions and determine
#         prediction intervals for predicted data
# **************************************************************

# Search parameters
scoring = "neg_mean_absolute_error"
cv = 2
n_jobs = -1
# n_iter = how many parameter combination to try in the search
n_iter = 10
random_state = 508842607

info = uq.quantile_regr_predint(
    x, y, jid, cv, n_jobs, n_iter, random_state, scoring, property
)

# TEST
assert info["MAE_Lower"] < 200.0
assert info["MAE_Mid"] < 200.0
assert info["MAE_Upper"] < 200.0
assert info["MAE_Error"] < 200.0
cmd ='rm *.pk *.dat'
os.system(cmd)



"""Module for showing periodic table chemical trends."""

from bokeh.models import (
    ColumnDataSource,
    LinearColorMapper,
    LogColorMapper,
    ColorBar,
    BasicTicker,
)
from bokeh.plotting import figure
from bokeh.io import output_file, save, show
from bokeh.sampledata.periodic_table import elements
from bokeh.transform import dodge
from matplotlib.colors import Normalize, LogNorm, to_hex
from matplotlib.cm import plasma, ScalarMappable

# , inferno, magma, viridis, ScalarMappable

# Reuires bokeh installation https://docs.bokeh.org/en/latest/


def plot_ptable_trend(
    data_elements=["Rb", "S", "Se"],
    data_list=[10, 20, 30],
    input_file=None,
    output_html="ptable.html",
    bokeh_palette="Plasma256",
    cmap=plasma,
    log_scale=0,
    width=1050,
    alpha=0.65,
    cbar_height=520,
    cbar_font="14pt",
    save_plot=True,
):
    """
    Generate periodic table chemical trends.

    Either provide a file or list of data_elements, &data_list.
    Note that Bokeh already provided a periodic table.
    This module will take your data to color code them.
    See an example: https://www.nature.com/articles/s41598-019-45028-y
    Fig. 3
    Forked from https://github.com/arosen93/ptable_trends
    """
    output_file(output_html)
    # Define number of and groups
    period_label = ["1", "2", "3", "4", "5", "6", "7"]
    group_range = [str(x) for x in range(1, 19)]
    if input_file is not None:
        data_elements = []
        data_list = []

        f = open(input_file, "r")
        lines = f.read().splitlines()
        f.close()
        for i in lines:
            data_elements.append(i.split()[0])
            data_list.append(i.split()[1])

    data = [float(i) for i in data_list]

    if len(data) != len(data_elements):
        raise ValueError("Unequal number of atomic elements and data points")

    # lanthanides = [x.lower() for x in elements["symbol"][56:70].tolist()]
    # actinides = [x.lower() for x in elements["symbol"][88:102].tolist()]
    period_label.append("blank")
    period_label.append("La")
    period_label.append("Ac")

    count = 0
    for i in range(56, 70):
        elements.period[i] = "La"
        elements.group[i] = str(count + 4)
        count += 1

    count = 0
    for i in range(88, 102):
        elements.period[i] = "Ac"
        elements.group[i] = str(count + 4)
        count += 1

    # Define matplotlib and bokeh color map
    if log_scale == 0:
        color_mapper = LinearColorMapper(
            palette=bokeh_palette, low=min(data), high=max(data)
        )
        norm = Normalize(vmin=min(data), vmax=max(data))
    elif log_scale == 1:
        for i in range(len(data)):
            if data[i] < 0:
                raise ValueError(
                    "Entry for element "
                    + data_elements[i]
                    + " is negative but"
                    " log-scale is selected"
                )
        color_mapper = LogColorMapper(
            palette=bokeh_palette, low=min(data), high=max(data)
        )
        norm = LogNorm(vmin=min(data), vmax=max(data))
    color_scale = ScalarMappable(norm=norm, cmap=cmap).to_rgba(
        data, alpha=None
    )

    # Define color for blank entries
    blank_color = "#c4c4c4"
    color_list = []
    for i in range(len(elements)):
        color_list.append(blank_color)

    # Compare elements in dataset with elements in periodic table
    for i in range(len(data)):
        element_entry = elements.symbol[
            elements.symbol.str.lower() == data_elements[i].lower()
        ]
        if not element_entry.empty:
            element_index = element_entry.index[0]
        else:
            print("WARNING: Invalid chemical symbol: " + data_elements[i])
        if color_list[element_index] != blank_color:
            print("WARNING: Multiple entries for element " + data_elements[i])
        color_list[element_index] = to_hex(color_scale[i])

    # Define figure properties for visualizing data
    source = ColumnDataSource(
        data=dict(
            group=[str(x) for x in elements["group"]],
            period=[str(y) for y in elements["period"]],
            sym=elements["symbol"],
            atomic_number=elements["atomic number"],
            type_color=color_list,
        )
    )
    # Plot the periodic table
    p = figure(
        x_range=group_range, y_range=list(reversed(period_label)), tools="save"
    )
    p.plot_width = width
    p.outline_line_color = None
    p.toolbar_location = "above"
    p.rect(
        "group",
        "period",
        0.9,
        0.9,
        source=source,
        alpha=alpha,
        color="type_color",
    )
    p.axis.visible = False
    text_props = {
        "source": source,
        "angle": 0,
        "color": "black",
        "text_align": "left",
        "text_baseline": "middle",
    }
    x = dodge("group", -0.4, range=p.x_range)
    y = dodge("period", 0.3, range=p.y_range)
    p.text(
        x=x,
        y="period",
        text="sym",
        text_font_style="bold",
        text_font_size="15pt",
        **text_props
    )
    p.text(x=x, y=y, text="atomic_number", text_font_size="9pt", **text_props)
    color_bar = ColorBar(
        color_mapper=color_mapper,
        ticker=BasicTicker(desired_num_ticks=10),
        border_line_color=None,
        label_standoff=6,
        major_label_text_font_size=cbar_font,
        location=(0, 0),
        orientation="vertical",
        scale_alpha=alpha,
        width=8,
    )

    if cbar_height is not None:
        color_bar.height = cbar_height

    p.add_layout(color_bar, "right")
    p.grid.grid_line_color = None
    if save_plot:
        save(p)
    else:
        show(p)
    return p


"""
if __name__ == "__main__":
    x = plot_ptable_trend()
    from bokeh.io import export_svgs
    # For svg
    x.output_backend = "svg"
    export_svgs(x, filename="plot.svg")
"""



"""Modules handling input files for VASP calculations."""
import pprint
import json
import os
import numpy as np
from jarvis.core.atoms import Atoms
from collections import OrderedDict
from jarvis.core.kpoints import generate_kgrid, Kpoints3D
from jarvis.core.utils import get_counts
from jarvis.core.specie import Specie
from jarvis.core.utils import update_dict


class Poscar(object):
    """
    Class defining Poscar object.

    Constructued from the Atoms object.

    Args:
        atoms : Atoms object

        comment : Header of Poscar file
    """

    def __init__(self, atoms, comment="System"):
        """Initialize the Poscar object."""
        self.atoms = atoms
        self.comment = comment

    @staticmethod
    def from_file(filename="POSCAR"):
        """Read simple POSCAR file from the path."""
        with open(filename, "r") as f:
            return Poscar.from_string(f.read())

    def to_dict(self):
        """Convert Poscar object to a dictionary."""
        d = OrderedDict()
        d["atoms"] = self.atoms.to_dict()
        d["comment"] = self.comment
        return d

    @classmethod
    def from_dict(self, d={}):
        """Construct Poscar object from a dictionary."""
        return Poscar(atoms=Atoms.from_dict(d["atoms"]), comment=d["comment"])

    def write_file(self, filename):
        """Write the Poscar object to a file."""
        f = open(filename, "w")
        header = (
            str(self.comment)
            + str("\n1.0\n")
            + str(self.atoms.lattice_mat[0][0])
            + " "
            + str(self.atoms.lattice_mat[0][1])
            + " "
            + str(self.atoms.lattice_mat[0][2])
            + "\n"
            + str(self.atoms.lattice_mat[1][0])
            + " "
            + str(self.atoms.lattice_mat[1][1])
            + " "
            + str(self.atoms.lattice_mat[1][2])
            + "\n"
            + str(self.atoms.lattice_mat[2][0])
            + " "
            + str(self.atoms.lattice_mat[2][1])
            + " "
            + str(self.atoms.lattice_mat[2][2])
            + "\n"
        )
        # order = np.argsort(self.atoms.elements)
        coords = self.atoms.frac_coords
        # DO NOT USE ORDER
        coords_ordered = np.array(coords)  # [order]
        elements_ordered = np.array(self.atoms.elements)  # [order]
        props_ordered = np.array(self.atoms.props)  # [order]
        # check_selective_dynamics = False
        counts = get_counts(elements_ordered)
        if "T" in "".join(map(str, self.atoms.props[0])):
            middle = (
                " ".join(map(str, counts.keys()))
                + "\n"
                + " ".join(map(str, counts.values()))
                + "\nSelective dynamics\n"
                + "Direct\n"
            )
        else:
            middle = (
                " ".join(map(str, counts.keys()))
                + "\n"
                + " ".join(map(str, counts.values()))
                + "\ndirect\n"
            )
        rest = ""
        # print ('repr',self.frac_coords, self.frac_coords.shape)
        for ii, i in enumerate(coords_ordered):
            p_ordered = str(props_ordered[ii])
            rest = rest + " ".join(map(str, i)) + " " + str(p_ordered) + "\n"

        result = header + middle + rest

        f.write(result)
        f.close()

    @staticmethod
    def from_string(lines):
        """Read Poscar from strings, useful in reading files/streams."""
        text = lines.splitlines()
        comment = text[0]
        scale = float(text[1])
        lattice_mat = []
        lattice_mat.append([float(i) for i in text[2].split()])
        lattice_mat.append([float(i) for i in text[3].split()])
        lattice_mat.append([float(i) for i in text[4].split()])
        lattice_mat = scale * np.array(lattice_mat)

        uniq_elements = text[5].split()
        element_count = np.array([int(i) for i in text[6].split()])
        elements = []
        for i, ii in enumerate(element_count):
            for j in range(ii):
                elements.append(uniq_elements[i])
        cartesian = True
        if "d" in text[7] or "D" in text[7]:
            cartesian = False
        # print ('cartesian poscar=',cartesian,text[7])
        num_atoms = int(np.sum(element_count))
        coords = []
        for i in range(num_atoms):
            coords.append([float(i) for i in text[8 + i].split()[0:3]])
        coords = np.array(coords)
        atoms = Atoms(
            lattice_mat=lattice_mat,
            coords=coords,
            elements=elements,
            cartesian=cartesian,
        )
        # print (atoms)
        # formula = atoms.composition.formula

        return Poscar(atoms, comment=comment)

    def __repr__(self):
        """Represent the Poscar class."""
        header = (
            str(self.comment)
            + str("\n1.0\n")
            + str(self.atoms.lattice_mat[0][0])
            + " "
            + str(self.atoms.lattice_mat[0][1])
            + " "
            + str(self.atoms.lattice_mat[0][2])
            + "\n"
            + str(self.atoms.lattice_mat[1][0])
            + " "
            + str(self.atoms.lattice_mat[1][1])
            + " "
            + str(self.atoms.lattice_mat[1][2])
            + "\n"
            + str(self.atoms.lattice_mat[2][0])
            + " "
            + str(self.atoms.lattice_mat[2][1])
            + " "
            + str(self.atoms.lattice_mat[2][2])
            + "\n"
        )
        order = np.argsort(self.atoms.elements)
        coords = self.atoms.frac_coords
        coords_ordered = np.array(coords)[order]
        elements_ordered = np.array(self.atoms.elements)[order]
        props_ordered = np.array(self.atoms.props)[order]
        # check_selective_dynamics = False
        counts = get_counts(elements_ordered)
        if "T" in "".join(map(str, self.atoms.props[0])):
            middle = (
                " ".join(map(str, counts.keys()))
                + "\n"
                + " ".join(map(str, counts.values()))
                + "\nSelective dynamics\n"
                + "Direct\n"
            )
        else:
            middle = (
                " ".join(map(str, counts.keys()))
                + "\n"
                + " ".join(map(str, counts.values()))
                + "\ndirect\n"
            )
        rest = ""
        # print ('repr',self.frac_coords, self.frac_coords.shape)
        for ii, i in enumerate(coords_ordered):
            p_ordered = str(props_ordered[ii])
            rest = rest + " ".join(map(str, i)) + " " + str(p_ordered) + "\n"
        result = header + middle + rest
        return result


class Incar(object):
    """Make VASP INCAR files as python dictionary of keys and values."""

    def __init__(self, tags={}):
        """Initialize with a dictionary."""
        self._tags = tags

    @staticmethod
    def from_file(filename="INCAR"):
        """Read INCAR file."""
        with open(filename, "r") as f:
            return Incar.from_string(f.read())

    def update(self, d={}):
        """Provide the new tags as a dictionary to update Incar object."""
        # print("selftags1=", self._tags)
        if self._tags != {}:
            for i, j in self._tags.items():
                self._tags[i] = j  # .strip(' ')
        for i, j in d.items():
            self._tags[i] = j
        # print("selftags2=", self._tags)
        return Incar(self._tags)

    # def get(self, key="POTIM", temp=0.5):
    #    """Get the value of a key."""
    #    if key in list(self._tags.keys()):
    #        return self._tags[key]
    #    else:
    #        self._tags[key] = temp
    #        print("Setting the temp value to the key", temp, key)
    #        return self._tags[key]

    def to_dict(self):
        """Convert into dictionary."""
        return self._tags

    @classmethod
    def from_dict(self, data={}):
        """Construct from dictionary."""
        return Incar(tags=data)

    @staticmethod
    def from_string(lines):
        """Construct from string."""
        text = lines.splitlines()
        tags = OrderedDict()
        for i in text:
            if "=" in i:
                tmp = i.split("=")
                tags.setdefault(tmp[0].strip(" "), tmp[1].strip(" "))

        return Incar(tags=tags)

    def __repr__(self):
        """Representation of the class."""
        return str(self._tags)

    def write_file(self, filename):
        """Write Incar to a file."""
        tags = self._tags
        lines = ""
        for i, j in tags.items():
            lines = lines + str(i) + str("=") + str(j) + "\n"
        f = open(filename, "w")
        f.write(lines)
        f.close()


class IndividualPotcarData(object):
    """Class for individual POTCAR file handling."""

    def __init__(self, data={}):
        """Intialize with some key info."""
        self._data = data

    def from_string(lines):
        """Generate some of the contents in the POTCAR."""
        text = lines.splitlines()
        individual_data = OrderedDict()
        individual_data["header1"] = text[0]
        individual_data["header2"] = text[2]
        individual_data["VRHFIN"] = text[3].split("=")[1]
        individual_data["element"] = text[3].split("=")[1].split(":")[0]
        individual_data["LEXCH"] = text[4].split("=")[1].replace(" ", "")
        tmp = text[5].split("=")[1].split()[0].replace(" ", "")
        individual_data["EATOM"] = tmp
        individual_data["TITEL"] = text[7].split("=")[1].replace(" ", "")
        return IndividualPotcarData(data=individual_data)

    def from_file(filename="POTCAR"):
        """Read from file."""
        with open(filename, "r") as f:
            return IndividualPotcarData.from_string(f.read())

    def __repr__(self):
        """Reprent the class."""
        return pprint.pformat(self._data, indent=4)


class Potcar(object):
    """Construct muti-atoms Postcar."""

    def __init__(
        self,
        elements=[],
        pot_type="POT_GGA_PAW_PBE",
        pot_json_path="",
        potcar_strings=[],
    ):
        """
        Initialize the Potcar class.

        POTCAR file contains the pseudopotential for each
        atomic species used in the calculation.
        Args:
            elements: atomic elements.

            pot_type: Type of pseudopotential.
            Look for VASP provided PSPs.
            VASP_PSP_DIR should be in the path.

            pot_json_path: Path to dictionary of chemical elements along with
            their orbitals taken into conisideration, e.g. V_pv.

            potcar_strings: One can directly provide
            the above mentioned strings.
        """
        self._elements = elements
        self._pot_type = pot_type
        self._potcar_strings = potcar_strings
        self._pot_json_path = pot_json_path

        if self._potcar_strings == []:
            pot_json_file = str(
                os.path.join(os.path.dirname(__file__), "default_potcars.json")
            )
            self._pot_json_path = pot_json_file
            pot_json = open(pot_json_file, "r")
            pot_json_selected = json.load(pot_json)
            pot_json.close()
            potcar_strings = []  # OrderedDict()
            for i in self._elements:
                potcar_strings.append(pot_json_selected[i])
                # for j, k in pot_json_selected.items():
                #    if i == j:
                #        potcar_strings.setdefault(i, k)
            self._potcar_strings = potcar_strings
            #  print("self._elements", self._elements)
            #  print("self._potcar_strings", self._potcar_strings)
            if len(self._elements) != len(self._potcar_strings):
                raise ValueError(
                    "Number of elements is not same as potcar_strings",
                    self._elements,
                    self._potcar_strings.keys(),
                )
        else:
            pot_json = open(self._pot_json_path, "r")
            pot_json_selected = json.load(pot_json)
            pot_json.close()
            self._potcar_strings = potcar_strings
            if len(self._elements) != len(self._potcar_strings):
                msg = "Number of elements not same as potcar_strings"
                raise ValueError(msg)

    @classmethod
    def from_dict(self, d={}):
        """Build class from a dictionary."""
        return Potcar(
            elements=d["elements"],
            pot_type=d["pot_type"],
            pot_json_path=d["pot_json_path"],
            potcar_strings=d["potcar_strings"],
        )

    def to_dict(self):
        """Convert to a dictionary."""
        d = OrderedDict()
        d["elements"] = np.array(self._elements).tolist()
        d["pot_type"] = self._pot_type
        d["pot_json_path"] = self._pot_json_path
        d["potcar_strings"] = np.array(self._potcar_strings).tolist()
        return d

    def catenate_potcar_files(
        self, destination_filename="POTCAR", filenames=[]
    ):
        """Catenate potcars of sifferent elements."""
        with open(destination_filename, "w") as outfile:
            for fname in filenames:
                with open(fname) as infile:
                    for line in infile:
                        outfile.write(line)

    def list_potcar_files(self):
        """List POTCAR files."""
        pot_files = []
        vasp_dir = str(os.environ["VASP_PSP_DIR"])
        vasp_psp_dir = str(os.path.join(vasp_dir, self._pot_type))
        potcar_strings = self._potcar_strings
        for j in potcar_strings:
            tmp = os.path.join(vasp_psp_dir, j, "POTCAR")
            pot_files.append(tmp)
        return pot_files

    def write_file(self, filename="POTCAR"):
        """Write POTCAR file."""
        pot_files = self.list_potcar_files()
        self.catenate_potcar_files(
            destination_filename=filename, filenames=pot_files
        )

    def __repr__(self):
        """Represent Potcar."""
        return str(str(self._pot_type) + "\n" + str(self._potcar_strings))


class Kpoints(object):
    """Make VASP KPOINTS as object."""

    def __init__(self, filename=""):
        """Initialize Kpoints from filename else read from file-stream."""
        self.filename = filename
        if filename != "":
            f = open(self.filename, "r")
            self.lines = f.read().splitlines()
            f.close()
            self.kpoints = self.read(self.lines)

    @classmethod
    def read(self, lines):
        """Read from an open file."""
        if lines[1] == "0":
            return self.get_mesh_kp(lines=lines)
        elif "Reciprocal" in lines[2]:
            return self.get_ibz_kp(lines=lines)
        else:
            ValueError("K-point scheme is not implemented")

    def get_mesh_kp(lines=""):
        """Read Kpoints as grid."""
        grid = [int(i) for i in lines[3].split()]
        kpts = generate_kgrid(grid)
        kpts_cls = Kpoints3D(kpoints=np.array(kpts))
        return kpts_cls

    def get_ibz_kp(lines=""):
        """Read the Kpoints in the line-mode."""
        kp_labels = []
        all_kp = []
        kp_labels_points = []
        for ii, i in enumerate(lines):
            if ii > 2:
                tmp = i.split()
                k = [tmp[0], tmp[1], tmp[2]]
                all_kp.append(k)
                if len(tmp) == 5:
                    tmp = str("$") + str(tmp[4]) + str("$")
                    if len(kp_labels) == 0:
                        kp_labels.append(tmp)
                        kp_labels_points.append(ii - 3)
                    elif tmp != kp_labels[-1]:
                        kp_labels.append(tmp)
                        kp_labels_points.append(ii - 3)
        labels = []
        for i in range(len(all_kp)):
            labels.append("")
        for i, j in zip(kp_labels, kp_labels_points):
            labels[j] = i
        all_kp = np.array(all_kp, dtype="float")
        kpts_cls = Kpoints3D(
            kpoints=all_kp, labels=labels, kpoint_mode="linemode"
        )
        return kpts_cls


def find_ldau_magmom(
    atoms="",
    default_magmom=True,
    U=3.0,
    mag=5.0,
    amix=0.2,
    bmix=0.00001,
    amixmag=0.8,
    bmixmag=0.00001,
    lsorbit=False,
):
    """Get necessary INCAR tags for DFT+U calculations."""
    sps = atoms.uniq_species
    LDAUL = []
    LDAUU = []
    LDAUTYPE = 2
    lmix = 4
    for i in sps:
        el = Specie(i)
        el_u = 0
        el_l = -1
        if el.element_property("is_transition_metal"):
            el_u = U
            el_l = 2
        if el.element_property("is_actinoid") or el.element_property(
            "is_lanthanoid"
        ):
            el_u = U
            el_l = 3
            lmix = 6
        LDAUL.append(el_l)
        LDAUU.append(el_u)
    if 3 in LDAUL:
        LDAUTYPE = 3
    nat = atoms.num_atoms
    magmom = str(nat) + str("*") + str(mag)
    if lsorbit:
        magmom = ""
        tmp = " 0 0 " + str(mag)
        for i in range(0, nat):
            magmom = magmom + tmp
    info = {}
    info["LDAU"] = ".TRUE."
    info["LDAUTYPE"] = LDAUTYPE
    info["LDAUL"] = " ".join(str(m) for m in LDAUL)
    info["LDAUU"] = " ".join(str(m) for m in LDAUU)
    info["LDAUPRINT"] = 2
    info["LMAMIX"] = lmix
    if not default_magmom:
        info["MAGMOM"] = magmom
    info["AMIX"] = amix
    info["BMIX"] = bmix
    info["AMIX_MAG"] = amixmag
    info["BMIX_MAG"] = bmixmag
    return info


def add_ldau_incar(
    use_incar_dict={}, atoms=None, Uval=2, lsorbit=False, default_magmom=True
):
    """Add LDAU in incase, especially made for spillage calcs."""
    info_ldau = find_ldau_magmom(
        U=Uval, atoms=atoms, lsorbit=lsorbit, default_magmom=default_magmom
    )
    tmp = update_dict(use_incar_dict, info_ldau)
    use_incar_dict = tmp
    return use_incar_dict



from jarvis.core.utils import update_dict,recast_array_on_uniq_array_elements,stringdict_to_xml,array_to_string
def test_utils():
   info={'a':'b','c':'d'}
   sd=stringdict_to_xml(info)
   sd=stringdict_to_xml(info,enforce_string=True)
   ar=[1,2,3,4,5]
   sarr=array_to_string(ar)
   x={'x':1,'y':2,'z':3}
   y={'m':1,'n':2,'o':3}
   z=update_dict(x,y)



"""Module for processing elastic tensor."""
# Reference: https://doi.org/10.1103/PhysRevB.98.014107
import numpy as np
from collections import OrderedDict


class ElasticTensor(object):
    """Module for processing elastic tensor."""

    def __init__(self, et_tensor=[]):
        """Initialize class."""
        self.et_tensor = et_tensor

    @property
    def voigt_modulus(self):
        """Get Voigt modulus."""
        c = self.et_tensor
        Kv = float(
            (c[0][0] + c[1][1] + c[2][2]) + 2 * (c[0][1] + c[1][2] + c[2][0])
        ) / float(9)
        Gv = float(
            (c[0][0] + c[1][1] + c[2][2])
            - (c[0][1] + c[1][2] + c[2][0])
            + 3 * (c[3][3] + c[4][4] + c[5][5])
        ) / float(15)
        return [Kv, Gv]

    @property
    def compliance_tensor(self):
        """Get compliance."""
        return np.linalg.inv(self.et_tensor)

    @property
    def reuss_modulus(self):
        """Get Reuss modulus."""
        c = self.compliance_tensor
        Kr = 1 / float(
            (c[0][0] + c[1][1] + c[2][2]) + 2 * (c[0][1] + c[1][2] + c[2][0])
        )
        Gr = 15 / (
            4 * (c[0][0] + c[1][1] + c[2][2])
            - 4 * (c[0][1] + c[1][2] + c[2][0])
            + 3 * (c[3][3] + c[4][4] + c[5][5])
        )
        return [Kr, Gr]

    @property
    def average_modulus(self):
        """Get average modulus."""
        return (
            np.array(self.voigt_modulus) + np.array(self.reuss_modulus)
        ) / 2

    @property
    def poisson_ratio(self):
        """Get poisson's ratio."""
        k, g = self.average_modulus
        return (3 * k - 2 * g) / (6 * k + 2 * g)

    @property
    def universal_ansiotropy_ratio(self):
        """Get universal ansiotropy ratio."""
        Kv, Gv = self.voigt_modulus
        Kr, Gr = self.reuss_modulus
        return 5 * (Gv / Gr) + (Kv / Kr) - 6

    @property
    def youngs_modulus(self):
        """Get Youngs modulus."""
        k, g = self.average_modulus
        return 9e9 * k * g / (3 * k + g)

    def to_dict(self):
        """Get dictionary representation."""
        d = OrderedDict()
        d["voigt_bulk_modulus"] = self.voigt_modulus[0]
        d["voigt_shear_modulus"] = self.voigt_modulus[1]
        d["reuss_bulk_modulus"] = self.reuss_modulus[0]
        d["reuss_shear_modulus"] = self.reuss_modulus[1]
        d["poisson_ratio"] = self.poisson_ratio
        d["youngs_modulus"] = self.youngs_modulus
        d["universal_ansiotropy_ratio"] = self.universal_ansiotropy_ratio
        d["raw_et_tensor"] = self.et_tensor
        return d


"""
from jarvis.io.vasp.outputs import Vasprun,Outcar
o=Outcar('../../examples/vasp/SiOptb88/SiOptb88/MAIN-ELASTIC-bulk@mp_149/OUTCAR')
print (o.elastic_props())
p=Atoms.from_poscar('../../examples/vasp/SiOptb88/SiOptb88/MAIN-ELASTIC-bulk@mp_149/POSCAR')
et=ElasticTensor(o.elastic_props()['cij'])
#print (et.voigt_modulus) #[87.26666666666667, 63.28]
#print (et.reuss_modulus) #[87.26666666666665, 60.24546397096941]
#print (et.average_modulus) #[87.26666667 61.76273199]
#print (et.poisson_ratio) #0.21367500388646996
#print (et.universal_ansiotropy_ratio) #0.21367500388646996
#print ('j_vel',(1e9*(et.average_modulus[0])/p.density)**.5)
#print ('j_vel',(1e9*
(et.average_modulus[0]+4/3*et.average_modulus[1])/p.density)**.5)
k,g=et.average_modulus
mass_density=p.density
#print ((1e9 * (k + 4. / 3. * g) / mass_density/1.6605e3 ) ** 0.5)
print (et.to_dict())
from pymatgen.analysis.elasticity.elastic import ElasticTensor
et=ElasticTensor.from_voigt(o.elastic_props()['cij'])
from pymatgen.core.structure import Structure
pmg=Structure.from_file('../../examples/vasp/
SiOptb88/SiOptb88/MAIN-ELASTIC-bulk@mp_149/POSCAR')
#print (et.k_vrh,et.g_vrh)
#print (et.long_v(pmg))
#print ('density_j,density_p',p.density,pmg.density)
"""



"""
Module for obtaining webpage data.

All the webpages are formatted in XML format.
They are converted into python dict from which
appropriate keys and values can be obtained.
Use 'list_keys' functions to find types
of data stored in each XML document.
"""
import numpy as np
import os
import requests
import tempfile
from jarvis.core.utils import xml_to_dict


class Webpage(object):
    """Get content of JARVIS-DFT and JARVIS-FF webpages."""

    def __init__(self, data=[], jid=""):
        """Initialize the class."""
        # JVASP- for JARVIS-DFT IDs
        # JLMP- for JARVIS-FF IDs

        self.jid = jid

        if "JVASP-" in jid:
            url = (
                "https://www.ctcms.nist.gov/~knc6/static/JARVIS-DFT/"
                + jid
                + ".xml"
            )
        if "JLMP-" in jid:
            url = (
                "https://www.ctcms.nist.gov/~knc6/static/JARVIS-FF/"
                + jid
                + ".xml"
            )
        self.url = url
        self.data = data
        if self.data == []:
            data = self.to_dict()
            self.data = data

    def to_dict(self):
        """Get dictionary for webpages."""
        dat = requests.get(self.url).text
        fd, path = tempfile.mkstemp()
        with os.fdopen(fd, "w") as tmp:
            tmp.write(dat)
        data = xml_to_dict(path)
        return data

    def list_keys(self):
        """Get all main keys from webpages."""
        print("keys=", self.data["basic_info"].keys())
        return self.data["basic_info"].keys()

    def get_dft_electron_dos(self):
        """Get JARVIS-DFT MAIN-RELAX DOS."""
        return self.data["basic_info"]["main_relax_info"]["main_relax_dos"]

    def get_dft_phonon_dos(self):
        """Get JARVIS-DFT MAIN-ELAST finite-diff DOS."""
        return self.data["basic_info"]["main_elastic"]["main_elastic_info"]

    def get_dft_mbj_dielectric_function(self):
        """Get JARVIS-DFT TBmBJ dielectric function."""
        info = {}
        energies = np.array(
            self.data["basic_info"]["main_optics_mbj"]["main_optics_mbj_info"][
                "energies"
            ]
            .strip("'")
            .split(","),
            dtype="float",
        )
        real_part_xx = np.array(
            self.data["basic_info"]["main_optics_mbj"]["main_optics_mbj_info"][
                "real_1"
            ]
            .strip("'")
            .split(","),
            dtype="float",
        )
        real_part_yy = np.array(
            self.data["basic_info"]["main_optics_mbj"]["main_optics_mbj_info"][
                "real_2"
            ]
            .strip("'")
            .split(","),
            dtype="float",
        )
        real_part_zz = np.array(
            self.data["basic_info"]["main_optics_mbj"]["main_optics_mbj_info"][
                "real_3"
            ]
            .strip("'")
            .split(","),
            dtype="float",
        )
        real_part_xy = np.array(
            self.data["basic_info"]["main_optics_mbj"]["main_optics_mbj_info"][
                "real_4"
            ]
            .strip("'")
            .split(","),
            dtype="float",
        )
        real_part_yz = np.array(
            self.data["basic_info"]["main_optics_mbj"]["main_optics_mbj_info"][
                "real_5"
            ]
            .strip("'")
            .split(","),
            dtype="float",
        )
        real_part_zx = np.array(
            self.data["basic_info"]["main_optics_mbj"]["main_optics_mbj_info"][
                "real_6"
            ]
            .strip("'")
            .split(","),
            dtype="float",
        )
        imag_part_xx = np.array(
            self.data["basic_info"]["main_optics_mbj"]["main_optics_mbj_info"][
                "imag_1"
            ]
            .strip("'")
            .split(","),
            dtype="float",
        )
        imag_part_yy = np.array(
            self.data["basic_info"]["main_optics_mbj"]["main_optics_mbj_info"][
                "imag_2"
            ]
            .strip("'")
            .split(","),
            dtype="float",
        )
        imag_part_zz = np.array(
            self.data["basic_info"]["main_optics_mbj"]["main_optics_mbj_info"][
                "imag_3"
            ]
            .strip("'")
            .split(","),
            dtype="float",
        )
        imag_part_xy = np.array(
            self.data["basic_info"]["main_optics_mbj"]["main_optics_mbj_info"][
                "imag_4"
            ]
            .strip("'")
            .split(","),
            dtype="float",
        )
        imag_part_yz = np.array(
            self.data["basic_info"]["main_optics_mbj"]["main_optics_mbj_info"][
                "imag_5"
            ]
            .strip("'")
            .split(","),
            dtype="float",
        )
        imag_part_zx = np.array(
            self.data["basic_info"]["main_optics_mbj"]["main_optics_mbj_info"][
                "imag_6"
            ]
            .strip("'")
            .split(","),
            dtype="float",
        )
        info["energies"] = energies
        info["real_xx"] = real_part_xx
        info["real_yy"] = real_part_yy
        info["real_zz"] = real_part_zz
        info["real_xy"] = real_part_xy
        info["real_yz"] = real_part_yz
        info["real_zx"] = real_part_zx
        info["imag_xx"] = imag_part_xx
        info["imag_yy"] = imag_part_yy
        info["imag_zz"] = imag_part_zz
        info["imag_xy"] = imag_part_xy
        info["imag_yz"] = imag_part_yz
        info["imag_zx"] = imag_part_zx
        return info

    def get_dft_semilocal_dielectric_function(self):
        """Get JARVIS-DFT semilocal dielectric function."""
        # OptB88vdW/LDA/PBE

        info = {}
        energies = np.array(
            self.data["basic_info"]["main_optics_semilocal"][
                "main_optics_info"
            ]["energies"]
            .strip("'")
            .split(","),
            dtype="float",
        )
        real_part_xx = np.array(
            self.data["basic_info"]["main_optics_semilocal"][
                "main_optics_info"
            ]["real_1"]
            .strip("'")
            .split(","),
            dtype="float",
        )
        real_part_yy = np.array(
            self.data["basic_info"]["main_optics_semilocal"][
                "main_optics_info"
            ]["real_2"]
            .strip("'")
            .split(","),
            dtype="float",
        )
        real_part_zz = np.array(
            self.data["basic_info"]["main_optics_semilocal"][
                "main_optics_info"
            ]["real_3"]
            .strip("'")
            .split(","),
            dtype="float",
        )
        real_part_xy = np.array(
            self.data["basic_info"]["main_optics_semilocal"][
                "main_optics_info"
            ]["real_4"]
            .strip("'")
            .split(","),
            dtype="float",
        )
        real_part_yz = np.array(
            self.data["basic_info"]["main_optics_semilocal"][
                "main_optics_info"
            ]["real_5"]
            .strip("'")
            .split(","),
            dtype="float",
        )
        real_part_zx = np.array(
            self.data["basic_info"]["main_optics_semilocal"][
                "main_optics_info"
            ]["real_6"]
            .strip("'")
            .split(","),
            dtype="float",
        )
        imag_part_xx = np.array(
            self.data["basic_info"]["main_optics_semilocal"][
                "main_optics_info"
            ]["imag_1"]
            .strip("'")
            .split(","),
            dtype="float",
        )
        imag_part_yy = np.array(
            self.data["basic_info"]["main_optics_semilocal"][
                "main_optics_info"
            ]["imag_2"]
            .strip("'")
            .split(","),
            dtype="float",
        )
        imag_part_zz = np.array(
            self.data["basic_info"]["main_optics_semilocal"][
                "main_optics_info"
            ]["imag_3"]
            .strip("'")
            .split(","),
            dtype="float",
        )
        imag_part_xy = np.array(
            self.data["basic_info"]["main_optics_semilocal"][
                "main_optics_info"
            ]["imag_4"]
            .strip("'")
            .split(","),
            dtype="float",
        )
        imag_part_yz = np.array(
            self.data["basic_info"]["main_optics_semilocal"][
                "main_optics_info"
            ]["imag_5"]
            .strip("'")
            .split(","),
            dtype="float",
        )
        imag_part_zx = np.array(
            self.data["basic_info"]["main_optics_semilocal"][
                "main_optics_info"
            ]["imag_6"]
            .strip("'")
            .split(","),
            dtype="float",
        )

        info["energies"] = energies
        info["real_xx"] = real_part_xx
        info["real_yy"] = real_part_yy
        info["real_zz"] = real_part_zz
        info["real_xy"] = real_part_xy
        info["real_yz"] = real_part_yz
        info["real_zx"] = real_part_zx
        info["imag_xx"] = imag_part_xx
        info["imag_yy"] = imag_part_yy
        info["imag_zz"] = imag_part_zz
        info["imag_xy"] = imag_part_xy
        info["imag_yz"] = imag_part_yz
        info["imag_zx"] = imag_part_zx
        return info


"""
if __name__ == "__main__":
    w = Webpage(jid="JVASP-1002")
    info_mbj = w.get_dft_mbj_dielectric_function()
    # print(info_mbj)
    w.list_keys()
"""



"""Modules for LightGBM regression."""

from sklearn.model_selection import (
    train_test_split,
    RandomizedSearchCV,
)
from sklearn.preprocessing import StandardScaler
from sklearn.feature_selection import VarianceThreshold
from sklearn.pipeline import Pipeline
import lightgbm as lgb
from jarvis.ai.pkgs.utils import regr_scores
from collections import defaultdict
import numpy as np
import pickle
import joblib
import matplotlib.pyplot as plt
import scipy as sp


def regression(
    X=[],
    Y=[],
    jid=[],
    test_size=0.1,
    plot=False,
    preprocess=True,
    feature_importance=True,
    save_model=False,
    feat_names=[],
    model_name="my_model",
    config={},
):
    """Get generic regression model."""
    lgbm = lgb.LGBMRegressor(
        n_estimators=config["n_estimators"],
        learning_rate=config["learning_rate"],
        num_leaves=config["num_leaves"],
    )
    info = defaultdict()

    X_train, X_test, y_train, y_test, jid_train, jid_test = train_test_split(
        X, Y, jid, random_state=1, test_size=test_size
    )
    pipe = Pipeline(
        [
            ("stdscal", StandardScaler()),
            ("vart", VarianceThreshold(1e-4)),
            ("est", lgbm),
        ]
    )
    if preprocess:
        model = pipe
    else:
        model = lgbm
    model.fit(X_train, y_train)
    pred = model.predict(X_test)
    reg_sc = regr_scores(y_test, pred)
    info["reg_scores"] = reg_sc

    if feature_importance:
        imp_data = []
        info["imp_data"] = imp_data
        if not preprocess:
            feat_imp = model.feature_importances_
            feat_imp = 100 * np.array(
                [float(i) / float(np.sum(feat_imp)) for i in feat_imp]
            )
            for f in range(len(feat_imp)):
                imp_data.append([feat_imp[f], feat_names[f]])
        else:
            feat_imp = model.named_steps["est"].feature_importances_
            feat_imp = 100 * np.array(
                [float(i) / float(np.sum(feat_imp)) for i in feat_imp]
            )
            keep_indices = model.named_steps["vart"].get_support(indices=True)
            indices = np.argsort(feat_imp)[::-1]
            new_feat_imp = feat_imp[indices]
            new_indices = keep_indices[indices]
            for f in range(len(new_feat_imp)):
                imp_data.append([new_feat_imp[f], feat_names[new_indices[f]]])

    print(
        model_name,
        round(reg_sc["mae"], 3),
        round(reg_sc["rmse"], 3),
        round(reg_sc["r2"], 3),
    )
    if plot:
        plt.plot(
            reg_sc["pred"],
            reg_sc["test"],
            ".",
            label=str(type(model).__name__)[0:4],
        )
        plt.legend()
        plt.xlabel("DFT")
        plt.ylabel("ML")

    if save_model:
        pk = str(model_name) + str(".pk")
        jb = str(model_name) + str(".jb")
        # js = str(model_name )+str('.js')
        pickle.dump(model, open(pk, "wb"))
        joblib.dump(model, jb)
        # TODO: implemet something like sklearn-json
        # json.dump(model.get_params(), open(js, "w"))
    return info


default_param_dist = {
    # 'boosting_type': [ 'dart'],
    # 'boosting_type': ['gbdt', 'dart', 'rf'],
    # 'num_leaves': sp.stats.randint(2, 1001),
    # 'subsample_for_bin': sp.stats.randint(10, 1001),
    # 'min_split_gain': sp.stats.uniform(0, 5.0),
    # 'min_child_weight': sp.stats.uniform(1e-6, 1e-2),
    # 'reg_alpha': sp.stats.uniform(0, 1e-2),
    # 'reg_lambda': sp.stats.uniform(0, 1e-2),
    # 'tree_learner': ['data', 'feature', 'serial', 'voting' ],
    # 'application': ['regression_l1', 'regression_l2', 'regression'],
    # 'bagging_freq': sp.stats.randint(1, 11),
    # 'bagging_fraction': sp.stats.uniform(.1, 0.9),
    # 'feature_fraction': sp.stats.uniform(.1, 0.9),
    # 'learning_rate': sp.stats.uniform(1e-3, 0.9),
    # 'est__num_leaves': [2,8,16],
    # 'est__min_data_in_leaf': [1,2,4],
    # 'est__learning_rate': [0.005,0.01,0.1],
    # 'est__max_depth': [1,3,5], #sp.stats.randint(1, 501),
    # 'est__n_estimators': [num_iteration,2*num_iteration,5*num_iteration],
    # sp.stats.randint(100, 20001),
    # 'gpu_use_dp': [True, False],
    "est__min_data_in_leaf": sp.stats.randint(5, 20),
    "est__n_estimators": sp.stats.randint(500, 2000),
    "est__num_leaves": sp.stats.randint(100, 500),
    "est__max_depth": sp.stats.randint(8, 20),
    "est__learning_rate": sp.stats.uniform(5e-3, 0.5),
}


def get_lgbm(
    train_x,
    val_x,
    train_y,
    val_y,
    cv,
    n_jobs,
    scoring,
    n_iter,
    objective,
    alpha,
    random_state,
    param_dist=default_param_dist,
):
    """
    Train a lightgbm model.

    Args:

        train_x: samples used for trainiing

        val_x: validation set

        train_y: train targets

        val_y: validation targets

        cv: # of cross-validations

        n_jobs: for making the job parallel

        scoring: scoring function to use such as MAE

    Returns:
           Best estimator.
    """
    # Get converged boosting iterations with high learning rate,
    # MAE as the convergence crietria
    lgbm = lgb.LGBMRegressor(
        n_estimators=500,
        learning_rate=0.1,
        max_depth=5,
        num_leaves=100,
        objective=objective,
        # min_data_in_leaf=2,
        n_jobs=-1,
        alpha=alpha,
        random_state=random_state,
        verbose=-1,
    )

    lgbm.fit(
        train_x,
        train_y,
        eval_set=[(val_x, val_y)],
        eval_metric="mae",
        # eval_metric='l1',
        early_stopping_rounds=10,
    )
    num_iteration = lgbm.best_iteration_
    print("num_iteration", num_iteration)
    print("in randomsearch cv")
    # Generally thousands of randomized search for optimal parameters
    # learning rate and num_leaves are very important
    lgbm = lgb.LGBMRegressor(
        objective=objective,
        # device='gpu',
        # n_estimators=num_iteration,
        n_jobs=n_jobs,
        alpha=alpha,
        verbose=-1,
    )
    pipe = Pipeline(
        [
            ("stdscal", StandardScaler()),
            ("vart", VarianceThreshold(1e-4)),
            ("est", lgbm),
        ]
    )

    # n_iter=10
    # Increase n_iter for production runs.
    rscv = RandomizedSearchCV(
        estimator=pipe,
        param_distributions=param_dist,
        cv=cv,
        scoring=scoring,
        n_iter=n_iter,
        n_jobs=n_jobs,
        verbose=-1,
        random_state=random_state,
        refit=True,
    )
    model = rscv.fit(train_x, train_y)
    print("Best Score: ", model.best_score_)
    print("Best Params: ", model.best_params_)
    print("Best Estimator: ", model.best_estimator_)
    # return model.best_estimator_
    return model


def parameters_dict():
    """Give example optimized parameters."""
    parameters = {
        "optb88vdw_bandgap": {
            "n_estimators": 324,
            "learning_rate": 0.06414333047469417,
            "num_leaves": 31,
        },
        "mbj_bandgap": {
            "n_estimators": 210,
            "learning_rate": 0.04727272041771037,
            "num_leaves": 121,
        },
        "epsx": {
            "n_estimators": 139,
            "learning_rate": 0.10098329400041395,
            "num_leaves": 527,
        },
        "epsy": {
            "n_estimators": 161,
            "learning_rate": 0.264679564828344,
            "num_leaves": 29,
        },
        "epsz": {
            "n_estimators": 161,
            "learning_rate": 0.264679564828344,
            "num_leaves": 29,
        },
        "mepsx": {
            "n_estimators": 75,
            "learning_rate": 0.05374708509141705,
            "num_leaves": 242,
        },
        "mepsy": {
            "n_estimators": 120,
            "learning_rate": 0.12048289662270327,
            "num_leaves": 398,
        },
        "mepsz": {
            "n_estimators": 89,
            "learning_rate": 0.09718152788954888,
            "num_leaves": 938,
        },
        "encut": {
            "n_estimators": 376,
            "learning_rate": 0.08982089572506267,
            "num_leaves": 762,
        },
        "kpoint_length_unit": {
            "n_estimators": 236,
            "learning_rate": 0.03234907667844313,
            "num_leaves": 794,
        },
        "bulk_modulus_kv": {
            "n_estimators": 380,
            "learning_rate": 0.08621497083536021,
            "num_leaves": 202,
        },
        "shear_modulus_gv": {
            "n_estimators": 284,
            "learning_rate": 0.017555838240950795,
            "num_leaves": 579,
        },
        "formation_energy_peratom": {
            "n_estimators": 1170,
            "learning_rate": 0.15375236057119931,
            "num_leaves": 273,
        },
        "exfoliation_energy": {
            "n_estimators": 47,
            "learning_rate": 0.0734095239247927,
            "num_leaves": 326,
        },
        "max_ir_mode": {
            "n_estimators": 500,
            "learning_rate": 0.0734095239247927,
            "num_leaves": 100,
        },
    }
    return parameters


"""
if __name__ == "__main__":
    from jarvis.ai.pkgs.utils import get_ml_data
    from jarvis.ai.descriptors.cfid import feat_names
    property = "exfoliation_energy"
    property='formation_energy_peratom'
    params = parameters_dict()[property]
    print(params)
    X, Y, jid = get_ml_data(dataset="cfid_3d", ml_property=property)
    names = feat_names()
    info = regression(X=X, Y=Y, jid = jid, config=params, feat_names=names)
    print(info)
"""



from jarvis.analysis.structure.spacegroup import (
    Spacegroup3D,
    symmetrically_distinct_miller_indices,
    get_wyckoff_position_operators,
)
from jarvis.core.atoms import Atoms
from jarvis.io.vasp.inputs import Poscar
import os
from collections import defaultdict
from jarvis.db.jsonutils import loadjson

s1 = Poscar.from_file(
    os.path.join(os.path.dirname(__file__), "..", "defects", "POSCAR-667.vasp")
).atoms
s2 = Poscar.from_file(
    os.path.join(os.path.dirname(__file__), "..", "..", "io", "wannier", "POSCAR")
).atoms
s3 = Poscar.from_file(os.path.join(os.path.dirname(__file__), "POSCAR-tetragonal")).atoms
s4 = Poscar.from_file(os.path.join(os.path.dirname(__file__), "POSCAR-Cmcm")).atoms
s5 = Poscar.from_file(os.path.join(os.path.dirname(__file__), "POSCAR-Aem2")).atoms
s6 = Poscar.from_file(os.path.join(os.path.dirname(__file__), "POSCAR-C2m")).atoms
s7 = Poscar.from_file(os.path.join(os.path.dirname(__file__), "POSCAR-Pc")).atoms
s8 = Poscar.from_file(os.path.join(os.path.dirname(__file__), "POSCAR-P-1")).atoms
s9 = Poscar.from_file(os.path.join(os.path.dirname(__file__), "POSCAR-P21m")).atoms


def test_spg229():
    d = loadjson(os.path.join(os.path.dirname(__file__), "spg229.json"))
    for i in d:
        atoms = Atoms.from_dict(i["atoms"])
        spg = Spacegroup3D(atoms).space_group_number
        assert spg == i["spg_number"]


def test_spg():
    box = [[2.715, 2.715, 0], [0, 2.715, 2.715], [2.715, 0, 2.715]]
    coords = [[0, 0, 0], [0.25, 0.25, 0.25]]
    elements = ["Si", "Si"]
    Si = Atoms(lattice_mat=box, coords=coords, elements=elements)
    spg = Spacegroup3D(atoms=Si)  # .spacegroup_data()
    cvn = spg.conventional_standard_structure
    ml = symmetrically_distinct_miller_indices(max_index=3, cvn_atoms=cvn)
    # print ('ml',ml)
    x = get_wyckoff_position_operators(488)
    # print  (x['wyckoff'][0]['letter'])
    assert (
        spg.space_group_number,
        spg.space_group_symbol,
        cvn.num_atoms,
        ml[0][0],
        x["wyckoff"][0]["letter"],
    ) == (227, "Fd-3m", 8, 1, "l")
    spg = Spacegroup3D(atoms=s1)
    assert spg.space_group_number == 191
    print(spg.space_group_number)
    cvn = spg.conventional_standard_structure
    spg = Spacegroup3D(atoms=s2)
    assert spg.space_group_number == 166
    print(spg.space_group_number)
    cvn = spg.conventional_standard_structure
    spg = Spacegroup3D(atoms=s3)
    print(spg.space_group_number)
    cvn = spg.conventional_standard_structure
    assert spg.space_group_number == 139
    spg = Spacegroup3D(atoms=s4)
    print(spg.space_group_number)
    cvn = spg.conventional_standard_structure
    assert spg.space_group_number == 63

    spg = Spacegroup3D(atoms=s5)
    print(spg.space_group_number)
    cvn = spg.conventional_standard_structure
    assert spg.space_group_number == 39

    spg = Spacegroup3D(atoms=s6)
    print(spg.space_group_number)
    cvn = spg.conventional_standard_structure
    assert spg.space_group_number == 12

    spg = Spacegroup3D(atoms=s7)
    print(spg.space_group_number)
    cvn = spg.conventional_standard_structure
    assert spg.space_group_number == 7

    spg = Spacegroup3D(atoms=s8)
    print(spg.space_group_number)
    cvn = spg.conventional_standard_structure
    assert spg.space_group_number == 2

    spg = Spacegroup3D(atoms=s9)
    print(spg.space_group_number)
    cvn = spg.conventional_standard_structure
    assert spg.space_group_number == 11


def test_all_spgs():
    """Moved to K-points module because of duplication."""
    return True


# test_all_spgs()
# test_spg()



"""Module to make XML for JARVIS-REST-API schema."""
import os
import traceback
import glob
import time
from jarvis.core.atoms import Atoms
from jarvis.core.atoms import VacuumPadding
import yaml
from jarvis.core.atoms import get_supercell_dims
from jarvis.analysis.thermodynamics.energetics import get_twod_defect_energy
from jarvis.io.vasp.outputs import Oszicar, Vasprun, Outcar
from matplotlib.pyplot import imread
from jarvis.analysis.structure.spacegroup import Spacegroup3D
from jarvis.analysis.topological.spillage import Spillage
from jarvis.db.jsonutils import loadjson
from jarvis.analysis.phonon.ir import ir_intensity
from jarvis.analysis.structure.neighbors import NeighborsAnalysis
from jarvis.analysis.solarefficiency.solar import SolarEfficiency
from jarvis.analysis.stm.tersoff_hamann import TersoffHamannSTM
from jarvis.analysis.thermodynamics.energetics import form_enp
from jarvis.tasks.boltztrap.run import run_boltztrap
from jarvis.tasks.phonopy.run import run_phonopy
from jarvis.io.phonopy.outputs import bandstructure_plot
from jarvis.ai.pkgs.utils import get_ml_data
from jarvis.analysis.diffraction.xrd import XRD
import numpy as np
from jarvis.analysis.elastic.tensor import ElasticTensor
from jarvis.io.boltztrap.outputs import BoltzTrapOutput
from jarvis.io.vasp.outputs import parse_raman_dat
from jarvis.core.utils import array_to_string
from jarvis.core.utils import stringdict_to_xml
from jarvis.ai.descriptors.cfid import CFID
from jarvis.io.wannier.outputs import WannierHam
from jarvis.db.figshare import data

cfid_x, cfid_y, cfid_jids = get_ml_data()

figshare_dict = data("raw_files")

wait_time = 20


def get_cfid_descriptors(
    jid="JVASP-1002", atoms="", make_cfid=True, material_type="bulk"
):
    """Get CFID pre-computed descriptors for a JID."""
    for i, j in zip(cfid_x, cfid_jids):
        if j == jid:
            return i
    if make_cfid:
        print("Calculating CFID descriptors.")
        if material_type == "Mol":
            atoms = VacuumPadding(atoms, vacuum=20.0).get_effective_molecule()
        if "Layer" in material_type:
            atoms = VacuumPadding(atoms, vacuum=20.0).get_effective_2d_slab()
        cfid = CFID(atoms)
        return cfid.get_comp_descp().tolist()


def get_figshare_files(jid="JVASP-1067"):
    """Get Figshare download links."""
    mem = []
    line = ""
    for i, j in figshare_dict.items():
        for k in j:
            if isinstance(k, dict):
                jjid = k["name"].split("_")[0].split(".zip")[0]
                if jjid == jid:
                    info = {}
                    info["category"] = i
                    info["url"] = k["download_url"]
                    info["name"] = k["name"]
                    line += ",".join([i, k["name"], k["download_url"]]) + ";"
                    info["line"] = line
                    mem.append(info)
    return line


icsd_mp_dat = loadjson("/rk2/knc6/JARVIS-DFT/icsd_mp_dat.json")

# Number of ICSDs
# x=[]
# for i in icsd_mp_dat:
#    mpid=i['mpid']
#    tmp=icsd_mp(mpid)
#    if tmp!='':
#       print (mpid,tmp)
#       y=tmp.split(',')
#       for j in y:
#         x.append(j)
# print (len(x))


def icsd_mp(ref=""):
    """Get ICSD IDs given MPIDs."""
    ids = ""
    for i in icsd_mp_dat:
        if i["mpid"] == ref:
            if len(i["icsds"]) != 0:
                if isinstance(i["icsds"], list):
                    ids = ",".join(map(str, i["icsds"]))
                else:
                    ids = str(i["icsds"]) + ","
    return ids


def get_see_also(folder=""):
    """Generate see also link."""
    line = ""
    tag = folder.split("/")[-1].split("_")[0]
    if "mp-" in tag or "mvc-" in tag:
        mpid = tag
        line += "MP: https://www.materialsproject.org/materials/" + mpid + ";"
        icsd = icsd_mp(mpid)
        line += "ICSD: " + icsd + ";"
    return line


class VaspToApiXmlSchema(object):
    """Module to convert VASP data to XML schema for API."""

    def __init__(
        self,
        folder="",
        meta_data={
            "id_file": "JARVIS-ID",
            "tmp_id_file": "'JARVIS-ID'",
            "ehull": "",
            "data_source": "JARVIS-DFT-VASP",
            "tmp_data_source": "'JARVIS-DFT-VASP'",
            "material_type": "bulk",
            "tmp_material_type": "'bulk'",
            "exfoliation_energy": "",
        },
    ):
        """Initialize class."""
        self.folder = folder
        meta_data["material_type"] = "bulk"
        if "1L" in folder:
            meta_data["material_type"] = "SingleLayer"
        if "2L" in folder:
            meta_data["material_type"] = "BiLayer"
        if "3L" in folder:
            meta_data["material_type"] = "TriLayer"
        if "Mol-" in folder:
            meta_data["material_type"] = "Molecule"
        self.meta_data = meta_data

    def ebandstruct(self, vrun="", kp=""):
        """Get electronic bandstucre related data."""
        line = ""
        try:
            f = open(kp, "r")
            lines = f.read().splitlines()
            f.close()

            vrun = Vasprun(vrun)
            band_indir_gap = ""
            band_dir_gap = ""
            try:
                band_indir_gap = str(round(vrun.get_indir_gap[0], 2))
                band_dir_gap = str(round(vrun.get_dir_gap, 2))
            except Exception:
                print("Cannot get bandgap.", kp)
                pass
            line += "<band_indir_gap>" + band_indir_gap + "</band_indir_gap>"
            line += "<band_dir_gap>" + band_dir_gap + "</band_dir_gap>"

            fermi_velocities = ""
            try:
                fermi_velocities = ",".join(map(str, vrun.fermi_velocities[0]))
            except Exception:
                print("Cannot get the Fermi-velocity.", kp)
                pass
            line += (
                '<fermi_velocities>"'
                + fermi_velocities
                + '"</fermi_velocities>'
            )

            kp_labels = []
            kp_labels_points = []
            for ii, i in enumerate(lines):
                if ii > 2:
                    tmp = i.split()
                    if len(tmp) == 5:
                        tmp = str(tmp[4])
                        if len(kp_labels) == 0:
                            kp_labels.append(tmp)
                            kp_labels_points.append(ii - 3)
                        elif tmp != kp_labels[-1]:
                            kp_labels.append(tmp)
                            kp_labels_points.append(ii - 3)
            tmp = 0.0
            zero_efermi = True
            if zero_efermi:
                tmp = float(vrun.efermi)
            spin = 0
            up_bands_y = []
            up_bands_x = []
            for i, ii in enumerate(vrun.eigenvalues[spin][:, :, 0].T - tmp):
                y = ",".join(map(str, ii))
                x = ",".join(map(str, range(0, len(ii))))
                up_bands_y.append(y)
                up_bands_x.append(x)
            spin = 1
            down_bands_y = []
            down_bands_x = []
            for i, ii in enumerate(vrun.eigenvalues[spin][:, :, 0].T - tmp):
                y = ",".join(map(str, ii))
                x = ",".join(map(str, range(0, len(ii))))
                down_bands_y.append(y)
                down_bands_x.append(x)
            line += (
                '<kp_labels_points>"'
                + ",".join(map(str, kp_labels_points))
                + '"</kp_labels_points>'
            )
            line += (
                '<kp_labels>"'
                + ",".join(map(str, kp_labels))
                + '"</kp_labels>'
            )
            line += (
                '<spin_up_bands_x>"'
                + ";".join(map(str, up_bands_x))
                + '"</spin_up_bands_x>'
            )
            line += (
                '<spin_up_bands_y>"'
                + ";".join(map(str, up_bands_y))
                + '"</spin_up_bands_y>'
            )
            line += (
                '<spin_down_bands_x>"'
                + ";".join(map(str, down_bands_x))
                + '"</spin_down_bands_x>'
            )
            line += (
                '<spin_down_bands_y>"'
                + ";".join(map(str, down_bands_y))
                + '"</spin_down_bands_y>'
            )
        except Exception:
            print("Could not get bandstructure info,vrun,kp", vrun, kp)
            pass

        return line

    def encut_kp(self):
        """Get ENCUT, cut-off convergence data."""
        folder = self.folder
        os.chdir(folder)
        info = {}

        try:
            encut_files = []
            encut_values = []
            encut_based_energies = []
            for i in glob.glob("ENCUT*.json"):
                if "ENCUT" in i:
                    encut_values.append(
                        int(str(i.split("-")[-1]).split(".json")[0])
                    )
                    encut_files.append(i)
                    erun = os.path.join(
                        folder, i.split(".json")[0], "vasprun.xml"
                    )
                    energy = float(Vasprun(erun).final_energy)
                    encut_based_energies.append(energy)

            kplength_files = []
            kp_values = []
            kp_based_energies = []
            for i in glob.glob("KPOINT*.json"):
                if "KPOINT" in i:
                    kp_values.append(
                        int(str(i.split("-")[-1]).split(".json")[0])
                    )
                    kplength_files.append(i)
                    krun = os.path.join(
                        folder, i.split(".json")[0], "vasprun.xml"
                    )
                    energy = float(Vasprun(krun).final_energy)
                    kp_based_energies.append(energy)
            encut_values = np.array(encut_values)
            encut_based_energies = np.array(encut_based_energies)
            order = np.argsort(encut_values)
            encut_values = encut_values[order].tolist()
            encut_based_energies = encut_based_energies[order].tolist()
            kp_values = np.array(kp_values)
            kp_based_energies = np.array(kp_based_energies)
            order = np.argsort(kp_values)
            kp_values = kp_values[order].tolist()
            kp_based_energies = kp_based_energies[order].tolist()

            info = {}
            info["converged_encut"] = str(sorted(list(set(encut_values)))[-6])
            info["converged_kpoint_length"] = str(
                sorted(list(set(kp_values)))[-6]
            )
            info["kp_values"] = "'" + ",".join(map(str, kp_values)) + "'"
            info["kp_based_energies"] = (
                "'" + ",".join(map(str, kp_based_energies)) + "'"
            )
            info["encut_values"] = "'" + ",".join(map(str, encut_values)) + "'"
            info["encut_based_energies"] = (
                "'" + ",".join(map(str, encut_based_energies)) + "'"
            )
            self.encut_kp_info = info
        except Exception:
            print("Check the convergence folders are named properly")
            pass
        os.chdir(folder)

        return info

    def electronic_dos_info(self, vrun):
        """Get electronic dos information with , _ and ; seperators."""
        line = ""
        try:
            try:
                line = ""
                total_dos = vrun.total_dos
                for i, ii in enumerate(total_dos):
                    if i == 0:
                        line = (
                            line
                            + "<edos_energies>'"
                            + ",".join(map(str, np.array(ii) - vrun.efermi))
                            + "'</edos_energies>"
                        )
                    elif i == 1:
                        line += (
                            "<total_edos_up>'"
                            + ",".join(map(str, ii))
                            + "'</total_edos_up>"
                        )
                    elif i == 2:
                        line += (
                            "<total_edos_down>'"
                            + ",".join(map(str, ii))
                            + "'</total_edos_down>"
                        )
                # energies = np.array(total_dos[0])
            except Exception as exp:
                print("No total dos", exp)
                pass
            try:
                spdf_dos = vrun.get_spdf_dos()
                atom_dos = vrun.get_atom_resolved_dos()

                line += "<spdf_dos>"
                for i, j in spdf_dos.items():
                    line += (
                        "<"
                        + str(i)
                        + ">"
                        # + ">'"
                        + "'"
                        + ",".join(map(str, j))
                        + "'"
                        + "</"
                        # + "'</"
                        + str(i)
                        + ">"
                    )

                line += "</spdf_dos>"
                line += "<elemental_dos>"
                spin_up_info = atom_dos["spin_up_info"]
                spin_down_info = atom_dos["spin_down_info"]
                line += '<spin_up_info>"'
                for i, j in spin_up_info.items():
                    line += str(i) + str("_") + ",".join(map(str, j)) + ";"
                line += '"</spin_up_info>'

                line += '<spin_down_info>"'
                # line += '<spin_down_info>"'
                for i, j in spin_down_info.items():
                    line += str(i) + str("_") + ",".join(map(str, j)) + ";"
                line += '"</spin_down_info>'
                # line += '"</spin_down_info>'
                line += "</elemental_dos>"
            except Exception as exp:
                print("No spdf dos", exp)
                pass
            fermi_velocities = ""
            try:
                fermi_velocities = ",".join(map(str, vrun.fermi_velocities[0]))
            except Exception:
                print("Cannot get the dos Fermi-velocity.")
                pass
            line += (
                '<fermi_velocities>"'
                + fermi_velocities
                + '"</fermi_velocities>'
            )

        except Exception as exp:
            print("Cannot get DOS data", vrun, exp)
            pass
        return line

    def get_xrd(self, atoms):
        """Get XRD pattern."""
        line = ""
        try:
            two_thetas, d_hkls, intensities = XRD().simulate(atoms=atoms)
            line += (
                '<two_thetas>"'
                + array_to_string(two_thetas)
                + '"</two_thetas>'
            )
            line += '<d_hkls>"' + array_to_string(d_hkls) + '"</d_hkls>'
            line += (
                '<intensities>"'
                + array_to_string(intensities)
                + '"</intensities>'
            )
        except Exception as exp:
            print("Cannot gt XRD pattern.", exp)
            pass
        return line

    def electronic_band_struct(self, vasprun="", kpoints_file_path=""):
        """Get electronic_band_struct."""
        line = self.ebandstruct(vrun=vasprun, kp=kpoints_file_path)
        return line

    def main_band(self):
        """Get bandstructure from MAIN-BAND-*."""
        folder = self.folder
        os.chdir(folder)
        data = ""
        info = {}
        for i in glob.glob("MAIN-BAND*.json"):
            try:
                folder = i.split(".json")[0]
                vrun = os.getcwd() + "/" + folder + "/vasprun.xml"
                kp = os.getcwd() + "/" + folder + "/KPOINTS"
                data = self.electronic_band_struct(
                    vasprun=vrun, kpoints_file_path=kp
                )
            except Exception as exp:
                print("Cannot get ebands info", folder, exp)
                pass
        os.chdir(folder)
        info["main_bands_info"] = data
        return info

    def main_hse06_band(self):
        """Get bandstructure from MAIN-HSE-*."""
        folder = self.folder
        os.chdir(folder)
        data = ""
        info = {}
        for i in glob.glob("MAIN-HSE*.json"):
            try:
                folder = i.split(".json")[0]
                vrun = os.getcwd() + "/" + folder + "/vasprun.xml"
                kp = os.getcwd() + "/" + folder + "/KPOINTS"
                data = self.electronic_band_struct(
                    vasprun=vrun, kpoints_file_path=kp
                )
            except Exception as exp:
                print("Cannot get hseband info", folder, exp)
                pass
        os.chdir(folder)
        info["main_hse_bands_info"] = data
        return info

    def main_pbe0_band(self):
        """Get bandstructure from MAIN-PBE0-*."""
        folder = self.folder
        os.chdir(folder)
        data = ""
        info = {}
        for i in glob.glob("MAIN-PBE0*.json"):
            try:
                folder = i.split(".json")[0]
                vrun = os.getcwd() + "/" + folder + "/vasprun.xml"
                kp = os.getcwd() + "/" + folder + "/KPOINTS"
                data = self.electronic_band_struct(
                    vasprun=vrun, kpoints_file_path=kp
                )
            except Exception as exp:
                print("Cannot get ebands info", folder, exp)
                pass
        os.chdir(folder)
        info["main_pbe0_bands_info"] = data
        return info

    def get_wannier_lines(self, data={}, energy_tol=4, write_full_data=False):
        """Get data for wannier- quality check."""
        # Dense mesh
        line = ""
        mesh_vrun_x = []
        mesh_vrun_y = []
        line += (
            "<maxdiff_mesh>'"
            + str(round(data["infomesh"]["maxdiff"], 4))
            + "'</maxdiff_mesh>"
        )
        line += (
            "<maxdiff_bz>'"
            + str(round(data["infobz"]["maxdiff"], 4))
            + "'</maxdiff_bz>"
        )
        if write_full_data:
            for i, ii in enumerate(np.array(data["infomesh"]["eigs_vrun"]).T):
                y = ",".join(map(str, ii))
                x = ",".join(map(str, range(0, len(ii))))
                mesh_vrun_y.append(y)
                mesh_vrun_x.append(x)
            line += (
                '<mesh_vrun_x>"'
                + ";".join(map(str, mesh_vrun_x))
                + '"</mesh_vrun_x>'
            )
            line += (
                '<mesh_vrun_y>"'
                + ";".join(map(str, mesh_vrun_y))
                + '"</mesh_vrun_y>'
            )

            min_arr = []
            erange = [-energy_tol, energy_tol]
            dd = {}
            kpp = data["infomesh"]["eigs_vrun"]
            vasp = np.array(data["infomesh"]["eigs_vrun"])
            wann = np.array(data["infomesh"]["eigs_wan"])
            for k in range(len(kpp)):
                for n in wann[k]:
                    diff_arr = []
                    if n > erange[0] and n < erange[1]:
                        for v in vasp[k]:
                            diff = abs(n - v)
                            diff_arr.append(diff)
                    if diff_arr != []:
                        tmp = np.min(diff_arr)
                        dd.setdefault(n, tmp)
                        min_arr.append(tmp)

            line += (
                '<mesh_difference_values>"'
                + str(",".join(map(str, dd.values())))
                + '"</mesh_difference_values>'
            )
            line += (
                '<mesh_difference_keys>"'
                + str(",".join(map(str, dd.keys())))
                + '"</mesh_difference_keys>'
            )

            mesh_wann_x = []
            mesh_wann_y = []
            for i, ii in enumerate(np.array(data["infomesh"]["eigs_wan"]).T):
                y = ",".join(map(str, ii))
                x = ",".join(map(str, range(0, len(ii))))
                mesh_wann_y.append(y)
                mesh_wann_x.append(x)
            line += (
                '<mesh_wann_x>"'
                + ";".join(map(str, mesh_wann_x))
                + '"</mesh_wann_x>'
            )
            line += (
                '<mesh_wann_y>"'
                + ";".join(map(str, mesh_wann_y))
                + '"</mesh_wann_y>'
            )

            # High symmetry BZ k-points

            bz_vrun_x = []
            bz_vrun_y = []
            for i, ii in enumerate(np.array(data["infobz"]["eigs_vrun"]).T):
                y = ",".join(map(str, ii))
                x = ",".join(map(str, range(0, len(ii))))
                bz_vrun_y.append(y)
                bz_vrun_x.append(x)
            line += (
                '<bz_vrun_x>"'
                + ";".join(map(str, bz_vrun_x))
                + '"</bz_vrun_x>'
            )
            line += (
                '<bz_vrun_y>"'
                + ";".join(map(str, bz_vrun_y))
                + '"</bz_vrun_y>'
            )

            min_arr = []
            erange = [-energy_tol, energy_tol]
            dd = {}
            kpp = data["infobz"]["eigs_vrun"]
            vasp = np.array(data["infobz"]["eigs_vrun"])
            wann = np.array(data["infobz"]["eigs_wan"])
            for k in range(len(kpp)):
                for n in wann[k]:
                    diff_arr = []
                    if n > erange[0] and n < erange[1]:
                        for v in vasp[k]:
                            diff = abs(n - v)
                            diff_arr.append(diff)
                    if diff_arr != []:
                        tmp = np.min(diff_arr)
                        dd.setdefault(n, tmp)
                        min_arr.append(tmp)

            line += (
                '<bz_difference_values>"'
                + str(",".join(map(str, dd.values())))
                + '"</bz_difference_values>'
            )
            line += (
                '<bz_difference_keys>"'
                + str(",".join(map(str, dd.keys())))
                + '"</bz_difference_keys>'
            )
        return line

    def wannier_comparison_plot(self, energy_tol=4):
        """Compare Wannier and DFT data on mesh and BZ."""
        folder = self.folder
        os.chdir(folder)
        line = ""
        info = {}
        # Change to WANN
        for i in glob.glob("MAIN-WANN*"):
            try:
                folder = i
                wann_dat = os.getcwd() + "/" + folder + "/wannier90_hr.dat"
                if os.path.exists(wann_dat):
                    wann_json = os.getcwd() + "/" + "/wannier_comparison.json"
                    print("Checking wann", wann_json)
                    if os.path.exists(wann_json):
                        data = loadjson(wann_json)
                        line += self.get_wannier_lines(data=data, energy_tol=4)
                        print(
                            "Getting data from preexisiting wannier json.",
                            line,
                        )
                    else:
                        print("Running jarvis-wannier solver.")
                        wann_ham = WannierHam(filename=wann_dat)

                        for j in glob.glob("MAIN-*.json"):
                            if "SOCSCFBAND" in j:
                                bands_vrun = (
                                    os.getcwd()
                                    + "/"
                                    + j.split(".json")[0]
                                    + "/vasprun.xml"
                                )
                            if "SOC" in j and "BAND" not in j:
                                mesh_vrun = (
                                    os.getcwd()
                                    + "/"
                                    + j.split(".json")[0]
                                    + "/vasprun.xml"
                                )

                        bz = wann_ham.compare_dft_wann(
                            vasprun_path=bands_vrun, plot=False
                        )
                        mesh = wann_ham.compare_dft_wann(
                            vasprun_path=mesh_vrun, plot=False
                        )
                        data = {}
                        data["infobz"] = bz
                        data["infomesh"] = mesh
                        line += self.get_wannier_lines(data=data, energy_tol=4)

            except Exception as exp:
                print("Cannot get wannier data.", folder, exp)
                pass
        # print("tmpw1", line)
        info["wannier_band_comparison"] = line
        return info

    def vacancy_formation_optb88vdw(self):
        """Get vacancy formation energy data."""
        folder = self.folder
        os.chdir(folder)

        line = ""
        info = {}

        try:
            vac_path = os.path.join(folder, "MAIN-VACANCY")
            if os.path.exists(vac_path):
                for i in glob.glob(vac_path + "/JVASP*"):
                    vac_vrun_path = os.path.join(
                        folder, "MAIN-VACANCY", i, "vasprun.xml"
                    )
                    outcar = Outcar(
                        os.path.join(folder, "MAIN-VACANCY", i, "OUTCAR")
                    )
                    if outcar.converged:
                        vac_vrun = Vasprun(vac_vrun_path)
                        tmp = i.split("/")[-1].split("_")
                        jid = tmp[0]
                        atom = tmp[2]
                        Ef = get_twod_defect_energy(
                            vrun=vac_vrun, jid=jid, atom=atom
                        )
                        line += (
                            jid + "_" + atom + "_" + str(round(Ef, 3)) + ","
                        )
        except Exception as exp:
            print("Cannot get vacancy info.", folder, exp)
            pass
        info["vacancy_formation_energy"] = line
        os.chdir(folder)
        return info

    def loptics_optoelectronics(self, vrun=""):
        """Get optoelctronic data."""
        line = ""

        try:
            lvrun = Vasprun(vrun)
            reals, imags = lvrun.dielectric_loptics
            energies = reals[:, 0]
            line = ""
            line += (
                "<energies>'" + ",".join(map(str, energies)) + "'</energies>"
            )
            line += '<loptics_dielectric_constant>"'
            line += str(reals[:, 1][0]) + ","
            line += str(reals[:, 2][0]) + ","
            line += str(reals[:, 3][0]) + ","
            line += str(reals[:, 4][0]) + ","
            line += str(reals[:, 5][0]) + ","
            line += str(reals[:, 6][0])
            line += '"</loptics_dielectric_constant>'
            lopt = ""
            try:
                max_linopt_eps = max(
                    reals[:, 1][0],
                    reals[:, 2][0],
                    reals[:, 3][0],
                    reals[:, 4][0],
                    reals[:, 5][0],
                    reals[:, 6][0],
                )
                lopt = str(round(max_linopt_eps, 2))
            except Exception as exp:
                print("Error in dielectric_loptics", exp)
                pass

            line += "<max_linopt_eps>" + lopt + "</max_linopt_eps>"
            fermi_velocities = ""
            try:
                fermi_velocities = ",".join(
                    map(str, lvrun.fermi_velocities[0])
                )
            except Exception as exp:
                print("Cannot get the lepsilon Fermi-velocity.", vrun, exp)
                pass
            line += (
                '<fermi_velocities>"'
                + fermi_velocities
                + '"</fermi_velocities>'
            )
            for i in np.arange(1, reals.shape[1]):
                line += (
                    "<real_"
                    + str(i)
                    + ">'"
                    + ",".join(map(str, reals[:, i]))
                    + "'</real_"
                    + str(i)
                    + ">"
                )
                line += (
                    "<imag_"
                    + str(i)
                    + ">'"
                    + ",".join(map(str, imags[:, i]))
                    + "'</imag_"
                    + str(i)
                    + ">"
                )
        except Exception as exp:
            print("Cannot get loptics data", vrun, exp)
            pass
        eff_slme = ""
        eff_sq = ""
        dirgap = ""
        indirgap = ""
        try:
            dirgap = round(lvrun.get_dir_gap, 3)
            indirgap = round(lvrun.get_indir_gap[0], 3)
            en, abz = lvrun.avg_absorption_coefficient
            abz = abz * 100
            eff_slme = SolarEfficiency().slme(
                en, abz, indirgap, indirgap, plot_current_voltage=False
            )
            # print("SLME", 100 * eff)
            eff_sq = SolarEfficiency().calculate_SQ(indirgap)
            eff_slme = round(100 * eff_slme, 2)
            eff_sq = round(100 * eff_sq, 2)
        except Exception as exp:
            print("Cannot get solar data.", exp)
            pass
        line += "<solar_slme>" + str(eff_slme) + "</solar_slme>"
        line += "<solar_sq>" + str(eff_sq) + "</solar_sq>"
        line += "<opto_dir_gap>" + str(dirgap) + "</opto_dir_gap>"
        line += "<opto_indir_gap>" + str(indirgap) + "</opto_indir_gap>"

        return line

    def main_optics_semilocal(self):
        """Get optoelctronic data from MAIN-OPTICS*."""
        folder = self.folder
        os.chdir(folder)
        info = {}
        data = ""
        for i in glob.glob("MAIN-OPTICS*.json"):

            try:
                folder = i.split(".json")[0]
                vrun = os.getcwd() + "/" + folder + "/vasprun.xml"
                data = self.loptics_optoelectronics(vrun)
            except Exception as exp:
                print("Cannot get semilocal optics data", folder, exp)
                pass
        os.chdir(folder)
        info["main_optics_info"] = data
        self.main_optics_semilocal_info = data
        return info

    def main_optics_mbj(self):
        """Get optoelctronic data from MAIN-MBJ*."""
        folder = self.folder
        os.chdir(folder)
        info = {}
        data = ""
        for i in glob.glob("MAIN-MBJ*.json"):
            try:
                folder = i.split(".json")[0]
                vrun = os.getcwd() + "/" + folder + "/vasprun.xml"
                data = self.loptics_optoelectronics(vrun)
            except Exception as exp:
                print("Cannot get mbj optics data", folder, exp)
                pass
        os.chdir(folder)
        info["main_optics_mbj_info"] = data
        self.main_optics_mbj_info = data
        return info

    def spillage(
        self,
        soc_wf="MAIN-SOCSCFBAND-bulk@JVASP-1067_mp-541837/WAVECAR",
        nonsoc_wf="MAIN-MAGSCFBAND-bulk@JVASP-1067_mp-541837/WAVECAR",
        spl_json_existing="/rk2/knc6/DB/SPILLAGE/all_spillage.json",
        jid="",
    ):
        """Get spillage data."""
        line = ""

        try:
            sp = {}
            """
            print("soc_wf", soc_wf)
            tmp_jid = jid
            print("tmp_jid", tmp_jid)
            if tmp_jid != "":
                spl_dat = loadjson(spl_json_existing)
                for i in spl_dat:
                    if i["jid"] == tmp_jid:
                        sp = i
                        sp["kpoints"] = list(
                            [[0, 0, 0] for i in range(len(sp["spillage_k"]))]
                        )
            """
            if sp == {}:
                # As spillage calculation may take a long time
                # print("XYZ1")
                sp = Spillage(
                    wf_noso=nonsoc_wf, wf_so=soc_wf
                ).overlap_so_spinpol()
                # print("XYZ2")
            soc_bands = self.electronic_band_struct(
                vasprun=soc_wf.replace("WAVECAR", "vasprun.xml"),
                kpoints_file_path=(soc_wf.replace("WAVECAR", "KPOINTS")),
            )
            line += "<soc_bands>" + soc_bands + "</soc_bands>"
            nonsoc_bands = self.electronic_band_struct(
                vasprun=nonsoc_wf.replace("WAVECAR", "vasprun.xml"),
                kpoints_file_path=(nonsoc_wf.replace("WAVECAR", "KPOINTS")),
            )
            line += "<nonsoc_bands>" + nonsoc_bands + "</nonsoc_bands>"
            spillage_k = ",".join(map(str, sp["spillage_k"]))
            line += "<spillage_k>'" + spillage_k + "'</spillage_k>"
            spillage_kpoints = ",".join(
                [";".join(map(str, i)) for i in sp["kpoints"]]
            )
            line += (
                "<spillage_kpoints>'"
                + spillage_kpoints
                + "'</spillage_kpoints>"
            )
            max_spillage = round(max(sp["spillage_k"]), 3)
            line += "<max_spillage>" + str(max_spillage) + "</max_spillage>"
        except Exception as exp:
            print("Cannot get spillage info", soc_wf, nonsoc_wf, exp)
            pass
        return line

    def main_soc_spillage(self, jid=""):
        """Get spillage data from folder."""
        folder = self.folder
        os.chdir(folder)
        info = {}
        data = ""
        for i in glob.glob("MAIN-SOCSCFBAND*.json"):
            try:
                folder = i.split(".json")[0]
                soc_wf = os.getcwd() + "/" + folder + "/WAVECAR"
                nonsoc_wf = (
                    os.getcwd()
                    + "/"
                    + folder.replace("SOC", "MAG")
                    + "/WAVECAR"
                )
                data = self.spillage(
                    soc_wf=soc_wf, nonsoc_wf=nonsoc_wf, jid=jid
                )
            except Exception as exp:
                print("Cannot get spillage info", folder, exp)
                pass
        os.chdir(folder)
        info["main_spillage_info"] = data
        self.main_soc_spillage_info = info
        return info

    def basic_info(
        self,
        include_dos_info=True,
        include_neighbor_info=True,
        use_cfid_data=True,
    ):
        """Get data from MAIN-RELAX folder."""
        main_folder = self.folder
        os.chdir(main_folder)
        info = {}
        info["source_folder"] = str(self.folder)
        try:
            info["see_also"] = ""
            see_also_dat = get_see_also(str(self.folder))
            info["see_also"] = '"' + see_also_dat + '"'
        except Exception as exp:
            print("Cannot get seel also data", exp)
            pass
        info["tmp_source_folder"] = "'" + str(self.folder) + "'"
        for i, j in self.meta_data.items():
            info[i] = j
        id_file = self.meta_data["id_file"]
        for i in glob.glob("MAIN-RELAX*.json"):
            folder = i.split(".json")[0]
            main_vrun = os.getcwd() + "/" + folder + "/vasprun.xml"
            main_contcar = os.getcwd() + "/" + folder + "/CONTCAR"
            main_outcar = os.getcwd() + "/" + folder + "/OUTCAR"
            # vrun = Vasprun(main_vrun)
            # atoms = vrun.all_structures[-1]
            try:
                final_energy = "na"
                scf_indir_gap = "na"
                scf_dir_gap = "na"
                vrun = Vasprun(main_vrun)
                final_energy = vrun.final_energy
                scf_indir_gap = vrun.get_indir_gap[0]
                scf_dir_gap = vrun.get_dir_gap

            except Exception as exp:
                print("Error in vasprun.xml.", exp)
            try:
                oszicar = Oszicar(os.path.join(folder, "OSZICAR"))
                magmom = oszicar.magnetic_moment
                info["magmom"] = round(float(magmom), 3)
                if final_energy == "na":
                    final_energy = float(oszicar.ionic_steps[-1][4])
            except Exception as exp:
                print("Check oszicar", exp)
                pass
            try:
                atoms = Atoms.from_poscar(main_contcar)
                # atoms = vrun.all_structures[-1]
                info["XRD"] = self.get_xrd(atoms)
                tmp_path = os.path.join(folder, "..", id_file)
                f = open(tmp_path, "r")
                lines = f.read().splitlines()
                f.close()
                id = lines[0]
                info["id"] = id
                try:
                    info["download_files"] = ""
                    files_line = get_figshare_files(id)
                    info["download_files"] = '"' + files_line + '"'
                except Exception as exp:
                    print("Cannot get figshare files.", exp)
                    pass
                info["tmp_id"] = '"' + str(id) + '"'
                formula = atoms.composition.reduced_formula
                info["formula"] = formula
                info["tmp_formula"] = '"' + formula + '"'
                elements = ",".join(atoms.uniq_species)
                info["elements"] = elements
                info["tmp_elements"] = '"' + elements + '"'
                info["number_uniq_species"] = len(atoms.uniq_species)
                method = "OptB88vdW"
                try:
                    f = open(os.path.join(folder, "..", "FUNCTIONAL"), "r")
                    lines = f.read().splitlines()
                    f.close()
                    method = lines[0]
                except Exception:
                    print('Cannot find "FUNCTIONAL"', method)
                    pass

                if method == "PBEBO":
                    method = "OptB88vdW"
                if method == "PBEOR":
                    method = "OptPBEvdW"
                if method == "PBEMK":
                    method = "OptB86bvdW"
                if (
                    "PBEBO" not in info["source_folder"]
                    and method == "OptB88vdW"
                ):
                    print(
                        "Inconsistent description, check INCAR",
                        info["source_folder"],
                    )
                info["method"] = method
                info["tmp_method"] = '"' + method + '"'
                num_atoms = atoms.num_atoms
                fen = ""
                if method == "OptB88vdW":
                    try:
                        fen = form_enp(atoms=atoms, total_energy=final_energy)
                    except Exception as exp:
                        print("Cannot get formation energy.", exp)
                        pass
                info["formation_energy"] = fen
                # print ('fen=',fen)
                rel_en = final_energy / num_atoms
                info["relaxed_energy"] = round(rel_en, 5)
                spg = Spacegroup3D(atoms)
                prim_atoms = spg.primitive_atoms
                conv_atoms = spg.conventional_standard_structure
                spg_numb = spg.space_group_number
                info["spacegroup_number"] = spg_numb
                info["point_group_symbol"] = spg.point_group_symbol
                # spg_symb = spg.space_group_symbol
                info["spg_space_group_symbol"] = spg.space_group_symbol
                crys_system = spg.crystal_system
                info["crys_system"] = crys_system
                conv_params = conv_atoms.lattice.parameters
                a_conv = round(conv_params[0], 2)
                b_conv = round(conv_params[1], 2)
                c_conv = round(conv_params[2], 2)
                info["a_conv"] = a_conv
                info["b_conv"] = b_conv
                info["c_conv"] = c_conv
                alpha_conv = round(conv_params[3], 2)
                beta_conv = round(conv_params[4], 2)
                gamma_conv = round(conv_params[5], 2)
                info["alpha_conv"] = alpha_conv
                info["beta_conv"] = beta_conv
                info["gamma_conv"] = gamma_conv
                prim_params = prim_atoms.lattice.parameters
                a_prim = round(prim_params[0], 2)
                b_prim = round(prim_params[1], 2)
                c_prim = round(prim_params[2], 2)
                info["a_prim"] = a_prim
                info["b_prim"] = b_prim
                info["c_prim"] = c_prim
                alpha_prim = round(prim_params[3], 2)
                beta_prim = round(prim_params[4], 2)
                gamma_prim = round(prim_params[5], 2)
                info["alpha_prim"] = alpha_prim
                info["beta_prim"] = beta_prim
                info["gamma_prim"] = gamma_prim
                prim_natoms = prim_atoms.num_atoms
                conv_natoms = conv_atoms.num_atoms
                info["prim_natoms"] = prim_natoms
                info["conv_natoms"] = conv_natoms
                density = round(atoms.density, 3)
                volume = round(atoms.volume, 3)
                info["density"] = density
                info["magnetization"] = ""
                info["total_charge"] = ""
                try:
                    tmp_out = Outcar(main_outcar)
                    elms = atoms.elements
                    magt = tmp_out.magnetization(elements=elms)
                    totc = tmp_out.total_charge(elements=elms)

                    info["magnetization"] = (
                        '"' + ";".join([",".join(ii) for ii in magt]) + '"'
                    )
                    info["total_charge"] = (
                        '"' + ";".join([",".join(ii) for ii in totc]) + '"'
                    )
                except Exception as exp:
                    print("magnetization, chg", exp)
                    pass
                if scf_indir_gap == "na":
                    try:
                        tmp_out = Outcar(main_outcar)
                        scf_indir_gap = tmp_out.bandgap[0]
                    except Exception as exp:
                        print("Outcar bandgap issue", exp)
                        pass
                info["volume"] = volume
                packing_fr = ""
                try:
                    packing_fr = round(atoms.packing_fraction, 3)
                except Exception:
                    print("Cannot calculate packing_fr, check atomic_radii")
                    pass
                info["packing_fr"] = packing_fr

                dim = get_supercell_dims(conv_atoms)
                # dim=[1,1,1]
                xyz = (
                    conv_atoms.make_supercell_matrix(dim)
                    .center_around_origin()
                    .get_xyz_string
                )
                info["xyz"] = '"' + str(xyz).replace("\n", "\\n") + '"'
                info["poscar_conv"] = (
                    '"'
                    + str(
                        conv_atoms.make_supercell_matrix(dim).get_string()
                    ).replace("\n", "\\n")
                    + '"'
                )

                info["contcar"] = (
                    '"' + str(atoms.get_string()).replace("\n", "\\n") + '"'
                )
                info["cif"] = ""
                try:
                    cif_filename = os.path.join(os.getcwd(), "atoms.cif")
                    atoms.write_cif(filename="atoms.cif")
                    cif_f = open(cif_filename, "r")
                    cif_lines = cif_f.read()  # .splitlines()
                    cif_f.close()
                    os.remove(cif_filename)
                    info["cif"] = (
                        '"' + str(cif_lines).replace("\n", "\\n") + '"'
                    )
                except Exception as exp:
                    print("CIF exception", exp)
                    pass

                # info["xyz"] = '"' + str(xyz).replace("\n", "\\n") + '"'
                rdf_bins = np.arange(0.1, 10.2, 0.1)
                ang1_bins = np.arange(1, 181.0, 1)
                ang2_bins = np.arange(1, 181.0, 1)
                dhd_bins = np.arange(1, 181.0, 1)
                rdf_hist = []
                ang1_hist = []
                ang2_hist = []
                dhd_hist = []
                info["cfid_descs"] = ""
                # Use pre-computed CFID dataset with chemo-structural features
                if use_cfid_data:
                    try:
                        cfid_descs = get_cfid_descriptors(
                            jid=id,
                            atoms=atoms,
                            material_type=self.meta_data["material_type"],
                        )
                        if cfid_descs is not None:
                            include_neighbor_info = False
                            rdf_hist = cfid_descs[820:920]
                            ang1_hist = cfid_descs[920:1099]
                            ang2_hist = cfid_descs[1099:1278]
                            dhd_hist = cfid_descs[1278:1457]
                            info["cfid_descs"] = (
                                "'" + ",".join(map(str, cfid_descs)) + "'"
                            )
                    except Exception as exp:
                        print("Cannot obtain cfid dataset", id, exp)
                        pass
                if include_neighbor_info:
                    nbr = NeighborsAnalysis(atoms)
                    rdf_bins, rdf_hist, nn = nbr.get_rdf()
                    nbr = NeighborsAnalysis(atoms, max_cut=10.0)
                    ang1_hist, ang1_bins = nbr.ang_dist_first()
                    ang2_hist, ang2_bins = nbr.ang_dist_second()
                    dhd_hist, dhd_bins = nbr.get_ddf()
                info["rdf_bins"] = "'" + ",".join(map(str, rdf_bins)) + "'"
                info["rdf_hist"] = "'" + ",".join(map(str, rdf_hist)) + "'"

                info["ang1_bins"] = "'" + ",".join(map(str, ang1_bins)) + "'"
                info["ang1_hist"] = "'" + ",".join(map(str, ang1_hist)) + "'"
                info["ang2_bins"] = "'" + ",".join(map(str, ang2_bins)) + "'"
                info["ang2_hist"] = "'" + ",".join(map(str, ang2_hist)) + "'"
                info["dihedral_bins"] = (
                    "'" + ",".join(map(str, dhd_bins)) + "'"
                )
                info["dihedral_hist"] = (
                    "'" + ",".join(map(str, dhd_hist)) + "'"
                )
                try:
                    info["scf_indir_gap"] = round(scf_indir_gap, 3)
                    info["scf_dir_gap"] = round(scf_dir_gap, 3)
                except Exception as exp:
                    print("Error in bandgap update info.", exp)
                    pass
                if include_dos_info:
                    main_relax_dos = self.electronic_dos_info(vrun)
                    info["main_relax_dos"] = main_relax_dos
            except Exception as expt:
                print("Exception in basic info", expt, traceback.format_exc())
                pass
        os.chdir(main_folder)
        self.basic_info_dict = info
        return info

    def dfpt_related(self, vrun="", out=""):
        """Get DFPT data."""
        vrun = Vasprun(vrun)
        data = vrun.dfpt_data
        out = Outcar(out)
        line = ""
        line += (
            "<epsilon>'"
            + ",".join(
                [
                    ";".join(map(str, data["epsilon"]["epsilon"][:, i]))
                    for i in range(0, 3)
                ]
            )
            + "'</epsilon>"
        )
        line += (
            "<epsilon_rpa>'"
            + ",".join(map(str, data["epsilon"]["epsilon_rpa"].flatten()))
            + "'</epsilon_rpa>"
        )
        line += (
            "<epsilon_ion>'"
            + ",".join(map(str, data["epsilon"]["epsilon_ion"].flatten()))
            + "'</epsilon_ion>"
        )
        max_eps = np.max(np.abs(np.array(data["epsilon"]["epsilon"])))
        line += "<max_eps>" + str(max_eps) + "</max_eps>"
        born_charges = data["born_charges"]
        atoms = vrun.all_structures[-1]
        spgg = Spacegroup3D(atoms)
        wycs = spgg._dataset["wyckoffs"]
        # spg = str(spgg._dataset["number"])
        natoms = atoms.num_atoms
        combs = []
        line += "<born_effective_charge>'"
        for k in range(natoms):
            comb = str(atoms.elements[k]) + "," + str(wycs[k])
            if comb not in combs:
                p = (
                    str(atoms.elements[k])
                    + ","
                    + str(wycs[k])
                    + ","
                    + ",".join(map(str, born_charges[k].flatten()))
                    + ";"
                )
                line += p
                combs.append(comb)
        line += "'</born_effective_charge>"
        # print (line)
        # vrun_eigs = data["phonon_eigenvalues"]
        pza = out.piezoelectric_tensor[1]
        max_pza = np.max(np.abs(pza))
        line += (
            "<max_piezo_stress_coeff>"
            + str(max_pza)
            + "</max_piezo_stress_coeff>"
        )
        pz = ",".join([";".join(map(str, pza[:, i])) for i in range(0, 6)])
        line += (
            '<dfpt_piezoelectric_tensor>"'
            + pz
            + '"</dfpt_piezoelectric_tensor>'
        )

        phonon_eigenvalues = out.phonon_eigenvalues
        phonon_eigenvectors = data["phonon_eigenvectors"]
        masses = data["masses"]
        born_charges = data["born_charges"]
        x, y = ir_intensity(
            phonon_eigenvectors=phonon_eigenvectors,
            phonon_eigenvalues=phonon_eigenvalues,
            masses=masses,
            born_charges=born_charges,
        )

        line += "<ir_intensity>'"
        line += ",".join(map(str, x)) + ";" + ",".join(map(str, y))
        line += "'</ir_intensity>"

        xx, yy = ir_intensity(
            smoothen=False,
            phonon_eigenvectors=phonon_eigenvectors,
            phonon_eigenvalues=phonon_eigenvalues,
            masses=masses,
            born_charges=born_charges,
        )
        xx = np.array(xx)
        yy = np.array(yy)
        max_ir_mode = ""
        min_ir_mode = ""
        try:
            tol = 1e-5
            non_zero_modes = yy > tol
            max_ir_mode = max(yy[non_zero_modes])
            min_ir_mode = min(yy[non_zero_modes])
        except Exception:
            print("Cannot find max min IR mode.", out)
        line += "<max_ir_mode>" + str(round(max_ir_mode, 2)) + "</max_ir_mode>"
        line += "<min_ir_mode>" + str(round(min_ir_mode, 2)) + "</min_ir_mode>"

        return line

    def main_lepsilon(self):
        """Get DFPT data from folder."""
        folder = self.folder
        os.chdir(folder)
        info = {}
        data = ""
        for i in glob.glob("MAIN-LEPSILON*.json"):
            try:
                folder = i.split(".json")[0]
                vrun = os.getcwd() + "/" + folder + "/vasprun.xml"
                out = os.getcwd() + "/" + folder + "/OUTCAR"
                data = self.dfpt_related(out=out, vrun=vrun)
            except Exception:
                print("Cannot find LEPSION data", folder)
                pass
        os.chdir(folder)
        info["main_lepsilon_info"] = data
        self.main_lepsilon_info = info
        return info

    def elastic_props(
        self, outcar="MAIN-ELASTIC-bulk@mp-149/OUTCAR",
    ):
        """Get elastic property data."""
        out = Outcar(outcar)
        try:
            out_phonon_eigenvalues = out.phonon_eigenvalues
        except Exception:
            print("Cannot get Outcar phonons.")
            pass
        vacuum = False
        atoms = None
        unit_system = "GPa"
        if "Layer" in self.meta_data["material_type"]:
            vacuum = True
            atoms = Atoms.from_poscar(outcar.replace("OUTCAR", "POSCAR"))
            unit_system = "Nm^-1"
        cij = np.array(out.elastic_props(atoms=atoms, vacuum=vacuum)["cij"])
        d = ElasticTensor(cij).to_dict()
        if "Layer" in self.meta_data["material_type"]:
            d = {}
        line = ""
        line += (
            '<cij>"'
            + ";".join([",".join(map(str, cij[:, i])) for i in range(0, 6)])
            + '"</cij>'
        )
        # print("line", line)
        for i, j in d.items():
            if i != "raw_et_tensor":
                line += (
                    "<" + str(i) + ">" + str(round(j, 2)) + "</" + str(i) + ">"
                )
        totdos = outcar.replace("OUTCAR", "total_dos.dat")
        line += '<unit_system>"' + unit_system + '"</unit_system>'
        cwd = str(os.getcwd())
        if not os.path.isfile(totdos):
            run_phonopy(outcar.split("/OUTCAR")[0])
        os.chdir(cwd)
        if os.path.isfile(totdos):

            mesh_yaml = outcar.replace("OUTCAR", "mesh.yaml")
            band_yaml = outcar.replace("OUTCAR", "band.yaml")
            with open(mesh_yaml, "r") as f:
                doc = yaml.load(f)
            nmodes = doc["phonon"][0]["band"]
            ph_modes = []

            for p in nmodes:
                ph_modes.append(p["frequency"])
            ph_modes = sorted(set(ph_modes))
            f = open(totdos, "r")
            freq = []
            pdos = []
            for lines in f.readlines():
                if not str(lines.split()[0]).startswith("#"):
                    #   print (lines)
                    # else:
                    freq.append(float(lines.split()[0]))
                    pdos.append(float(lines.split()[1]))
            line += (
                "<phonon_modes>"
                + ",".join(map(str, [round(i, 2) for i in ph_modes]))
                + "</phonon_modes>"
            )
            line += (
                "<min_fd_phonon_mode>"
                + str(round(min(out_phonon_eigenvalues), 1))
                + "</min_fd_phonon_mode>"
            )
            line + '<outcar_phonons>"' + ",".join(
                map(str, out_phonon_eigenvalues)
            ) + '"</outcar_phonons>'
            line += (
                "<phonon_dos_frequencies>'"
                + ",".join(map(str, freq))
                + "'</phonon_dos_frequencies>"
            )
            line += (
                "<phonon_dos_intensity>'"
                + ",".join(map(str, pdos))
                + "'</phonon_dos_intensity>"
            )

            # Comment until 4 MB text size error
            frequencies, distances, labels, label_points = bandstructure_plot(
                band_yaml
            )
            tmp = ""
            for i in range(np.array(frequencies).shape[1]):
                tmp += ",".join(map(str, np.array(frequencies)[:, i])) + ";"
            line += (
                "<phonon_bandstructure_distances>'"
                + ",".join(map(str, distances))
                + "'</phonon_bandstructure_distances>"
            )
            line += (
                "<phonon_bandstructure_frequencies>'"
                + tmp
                + "'</phonon_bandstructure_frequencies>"
            )
            line += (
                "<phonon_bandstructure_labels>'"
                + ",".join(map(str, labels))
                + "'</phonon_bandstructure_labels>"
            )
            line += (
                "<phonon_bandstructure_label_points>'"
                + ",".join(map(str, label_points))
                + "'</phonon_bandstructure_label_points>"
            )
        return line

    def main_elastic(self):
        """Get elastic property data from folder."""
        folder = self.folder
        os.chdir(folder)
        info = {}
        data = ""

        for i in glob.glob("MAIN-ELAST*.json"):

            try:
                folder = i.split(".json")[0]
                out = os.getcwd() + "/" + folder + "/OUTCAR"
                data = self.elastic_props(out)
            except Exception:
                print("Cannot get ELAST folder", folder)
                pass
        os.chdir(folder)
        info["main_elastic_info"] = data
        self.main_elastic_info = info
        return info

    def raman_data(self):
        """Get Raman intensity."""
        folder = self.folder
        os.chdir(folder)
        info = {}
        line = ""

        for i in glob.glob("RAMANDIR*/vasp_raman.dat"):
            try:
                raman_file = os.path.join(os.getcwd(), i)
                ram = open(raman_file, "r")
                lines = ram.read().splitlines()
                ram.close()
                an = raman_file.replace("vasp_raman.dat", "POSCAR")
                pos = Atoms.from_poscar(an)
                if len(lines) == 3 * pos.num_atoms + 1:
                    data = parse_raman_dat(raman_file)
                    line += (
                        '<frequencies>"'
                        + array_to_string(data["freqs"])
                        + '"</frequencies>'
                    )
                    line += (
                        '<activity>"'
                        + array_to_string(data["activity"])
                        + '"</activity>'
                    )
                    line += (
                        '<alpha>"'
                        + array_to_string(data["alpha"])
                        + '"</alpha>'
                    )
                    line += (
                        '<beta2>"'
                        + array_to_string(data["beta2"])
                        + '"</beta2>'
                    )
                    line += (
                        '<indices>"'
                        + array_to_string(data["beta2"])
                        + '"</indices>'
                    )
                    line += (
                        "<max_raman_mode>"
                        + str(round(max(data["freqs"]), 2))
                        + "</max_raman_mode>"
                    )
            except Exception:
                print("Cannot process Raman data.")
                pass
        os.chdir(folder)
        info["raman_dat"] = line
        return info

    def effective_mass_data(self,):
        """Get effective-mass data."""
        folder = self.folder
        info = {}
        line = ""
        os.chdir(folder)

        try:
            trap = os.path.join(folder, "trap_info.json")
            data = loadjson(trap)
            avg = data["avg_mass"]
            ndat = avg["n"]["300"][0]["data"]
            pdat = avg["p"]["300"][0]["data"]
            line += (
                '<electron_mass_300K>"'
                + str(round(ndat[0][0], 2))
                + ","
                + str(round(ndat[1][1], 2))
                + ","
                + str(round(ndat[2][2], 2))
                + '"</electron_mass_300K>'
            )
            line += (
                '<hole_mass_300K>"'
                + str(round(pdat[0][0], 2))
                + ","
                + str(round(pdat[1][1], 2))
                + ","
                + str(round(pdat[2][2], 2))
                + '"</hole_mass_300K>'
            )
            line += (
                "<max_electron_mass_300K>"
                + str(round(max(ndat[0][0], ndat[1][1], ndat[2][2]), 2))
                + "</max_electron_mass_300K>"
            )
            line += (
                "<max_hole_mass_300K>"
                + str(round(max(pdat[0][0], pdat[1][1], pdat[2][2]), 2))
                + "</max_hole_mass_300K>"
            )

        except Exception:
            print("Cannot get trap info.")
            pass
        os.chdir(folder)
        info["effective_mass"] = line
        return info

    def boltztrap_data(
        self,
        path="MAIN-RELAX-bulk@mp-149/boltztrap",
        temperature=600,
        doping=1e20,
    ):
        """Get transport data."""
        info = {}
        cwd = str(os.getcwd())
        if not os.path.exists(path):
            run_boltztrap(path.split("/boltztrap")[0])
        os.chdir(cwd)
        all_data = BoltzTrapOutput(path).to_dict()
        # info["all_data"] = all_data
        small_p = all_data["condtens_fixdoping"]["p"][temperature]
        small_n = all_data["condtens_fixdoping"]["n"][temperature]
        for i, j in small_p.items():
            if j["N_cm3"] == doping:
                tmp = j
                pseeb = np.array(
                    [
                        np.real(r)
                        for r in np.linalg.eigvals(
                            tmp["seeb"].reshape(3, 3) * 1e6
                        )
                    ]
                )
                pcond = np.array(
                    [
                        np.real(r)
                        for r in np.linalg.eigvals(tmp["cond"].reshape(3, 3))
                        / 1e14
                    ]
                )
                ppf = pseeb ** 2 * pcond / 1e6
                pkappa = np.linalg.eigvals(tmp["kappa"].reshape(3, 3))

                info["pseeb"] = [round(k, 2) for k in pseeb.tolist()]
                info["pcond"] = [round(k, 2) for k in pcond.tolist()]
                info["ppf"] = [round(k, 2) for k in ppf.tolist()]
                info["pkappa"] = [round(k, 2) for k in pkappa.tolist()]

        for i, j in small_n.items():
            if j["N_cm3"] == -1 * doping:
                tmp = j
                nseeb = np.array(
                    [
                        np.real(r)
                        for r in np.linalg.eigvals(
                            tmp["seeb"].reshape(3, 3) * 1e6
                        )
                    ]
                )
                ncond = np.array(
                    [
                        np.real(r)
                        for r in np.linalg.eigvals(tmp["cond"].reshape(3, 3))
                        / 1e14
                    ]
                )
                npf = nseeb ** 2 * ncond / 1e6
                nkappa = np.linalg.eigvals(tmp["kappa"].reshape(3, 3))

                info["nseeb"] = np.array(
                    [round(np.real(r), 2) for r in nseeb.tolist()]
                )
                info["ncond"] = np.array(
                    [round(np.real(r), 2) for r in ncond.tolist()]
                )
                info["npf"] = [round(k, 2) for k in npf.tolist()]
                info["nkappa"] = [round(k, 2) for k in nkappa.tolist()]
        line = ""
        for i, j in info.items():
            if isinstance(j, list):
                tmp = ";".join(map(str, j))
            else:
                tmp = j
            line += (
                "<"
                + str(i)
                + ">'"
                + ",".join(map(str, j))
                + "'</"
                + str(i)
                + ">"
            )
        # print (line)
        # print (all_data)
        return line

    def main_boltz_data(self):
        """Get transport data from folder."""
        folder = self.folder
        info = {}
        data = ""
        os.chdir(folder)
        for i in glob.glob("MAIN-RELAX*.json"):
            try:
                folder = i.split(".json")[0]
                path = os.getcwd() + "/" + folder + "/boltztrap"
                data = self.boltztrap_data(path=path)
            except Exception:
                print("Cannot find boltztrap folder", folder)
                pass
        os.chdir(folder)
        info["boltztrap_info"] = data
        self.boltztrap_info_dict = info
        return info

    def image_to_string(
        self, img_path="2DSTM/PNG_JSON3/JVASP-60776_mp-19795_pos.jpg",
    ):
        """Transform image to string."""
        # 2D array only
        fig = imread(img_path)
        line = (
            str(fig.shape[0])
            + "_"
            + str(fig.shape[1])
            + "_"
            + ",".join(map(str, fig[:, :, 0].flatten()))
        )
        # Comment when 4 MBlimit is over
        line = ""
        return line

    def main_stm_neg(self):
        """Get neg. bias constant height STM image."""
        folder = self.folder
        line = ""
        for i in glob.glob("MAIN-STM-NEG*.json"):
            try:
                folder = i.split(".json")[0]
                pchg = os.getcwd() + "/" + folder + "/PARCHG"
                TH_STM = TersoffHamannSTM(chg_name=pchg)
                filename = os.getcwd() + "/" + folder + "/testh.jpg"
                t1 = TH_STM.constant_height(filename=filename)
                zcut = t1["zcut"]
                print(zcut)
                # line += "<zcut>" + str(zcut) + "</zcut>"
                line += self.image_to_string(img_path=filename)
                cmd = "rm -rf " + filename
                if os.path.isfile(filename):
                    os.system(cmd)
            except Exception:
                print("Couldnt make negative bias STM image", folder)
                pass
        return line

    def main_stm_pos(self):
        """Get pos. bias constant height STM image."""
        folder = self.folder
        line = ""
        for i in glob.glob("MAIN-STM-POS*.json"):
            try:
                folder = i.split(".json")[0]
                pchg = os.getcwd() + "/" + folder + "/PARCHG"
                TH_STM = TersoffHamannSTM(chg_name=pchg)
                filename = os.getcwd() + "/" + folder + "/testh.jpg"
                t1 = TH_STM.constant_height(filename=filename)
                zcut = t1["zcut"]
                print(zcut)
                # line += "<zcut>" + str(zcut) + "</zcut>"
                line += self.image_to_string(img_path=filename)
                cmd = "rm -rf " + filename
                if os.path.isfile(filename):
                    os.system(cmd)
            except Exception:
                print("Couldnt make positive bias STM image", folder)
                pass
        return line

    def efg_tensor(self):
        """Get electric field gradient tensor."""
        main_folder = self.folder
        info = {}
        efg_dat = ""
        max_efg = ""
        os.chdir(main_folder)
        for i in glob.glob("MAIN-LEFG*.json"):
            try:
                folder = i.split(".json")[0]
                out = os.getcwd() + "/" + folder + "/OUTCAR"
                atoms = Atoms.from_poscar(out.replace("OUTCAR", "POSCAR"))
                out = Outcar(out)
                spgg = Spacegroup3D(atoms)
                wycs = spgg._dataset["wyckoffs"]
                natoms = atoms.num_atoms
                combs = []
                efg = out.efg_raw_tensor
                efg_dat = ""
                max_efg = np.max(np.abs(np.array(efg)))
                for k in range(natoms):
                    comb = str(atoms.elements[k]) + "," + str(wycs[k])
                    if comb not in combs:
                        efg_dat = (
                            efg_dat
                            + comb
                            + ","
                            + ",".join(map(str, efg[k].flatten()))
                            + ";"
                        )
                        combs.append(comb)
            except Exception:
                print("Cannot get LEFG folder", folder)
                pass

        info["efg_raw_tensor"] = efg_dat
        info["max_efg"] = max_efg
        info["max_efg_eta"] = ""
        # self.efg_raw_tensor_info = info
        os.chdir(main_folder)
        return info

    def write_xml(self, filename="temp.xml"):
        """Get overall XML file."""
        # f=open(filename,"w")
        line = '<?xml version="1.0" encoding="UTF-8"?>\n'
        line += '<?xml-stylesheet type="text/xsl" '
        line += 'href="jarvisdft.xsl"?>\n<basic_info>'
        # f.write(str(line))
        line += (
            "<convergence_info>"
            + stringdict_to_xml(self.encut_kp())
            + "</convergence_info>"
            + "\n"
        )
        # f.write(str(line))
        tmpw = self.wannier_comparison_plot()
        # print ('tmpw=',tmpw,stringdict_to_xml(tmpw))
        line += stringdict_to_xml(tmpw) + "\n"
        # f.write(str(line))
        print(line)
        line += stringdict_to_xml(self.vacancy_formation_optb88vdw()) + "\n"
        # f.write(str(line))
        line += stringdict_to_xml(self.raman_data()) + "\n"
        # f.write(str(line))
        b_info = self.basic_info()
        jid = b_info["id"]
        line += (
            "<main_relax_info>"
            + stringdict_to_xml(b_info)
            + "</main_relax_info>"
            + "\n"
        )
        # f.write(str(line))
        line += (
            "<main_band>"
            + stringdict_to_xml(self.main_band(), enforce_string=False)
            + "</main_band>"
        )
        # f.write(str(line))
        line += (
            "<main_hse06_band>"
            + stringdict_to_xml(self.main_hse06_band(), enforce_string=False)
            + "</main_hse06_band>"
        )
        # f.write(str(line))

        line += stringdict_to_xml(self.effective_mass_data()) + "\n"
        # f.write(str(line))

        line += (
            "<main_pbe0_band>"
            + stringdict_to_xml(self.main_pbe0_band(), enforce_string=False)
            + "</main_pbe0_band>"
        )
        # f.write(str(line))
        line += (
            "<main_optics_semilocal>"
            + stringdict_to_xml(self.main_optics_semilocal())
            + "</main_optics_semilocal>"
            + "\n"
        )
        # f.write(str(line))
        line += (
            "<main_optics_mbj>"
            + stringdict_to_xml(self.main_optics_mbj())
            + "</main_optics_mbj>"
            + "\n"
        )
        # f.write(str(line))
        line += (
            "<main_elastic>"
            + stringdict_to_xml(self.main_elastic())
            + "</main_elastic>"
            + "\n"
        )
        # f.write(str(line))
        line += (
            "<main_boltz>"
            + stringdict_to_xml(self.main_boltz_data())
            + "</main_boltz>"
            + "\n"
        )
        # f.write(str(line))
        line += stringdict_to_xml(self.main_lepsilon()) + "\n"
        # f.write(str(line))
        line += stringdict_to_xml(self.main_soc_spillage(jid=jid)) + "\n"
        # f.write(str(line))
        line += (
            stringdict_to_xml(self.efg_tensor(), enforce_string=True) + "\n"
        )
        # f.write(str(line))
        stm_neg = ""
        stm_pos = ""
        stm_pos_file = (
            "/u/WWW/knc6/static/JARVIS-DFT-STM/Images/JARVIS-2D-STM-JPG/"
            + str(jid)
            + "_pos.jpg"
        )
        stm_neg_file = (
            "/u/WWW/knc6/static/JARVIS-DFT-STM/Images/JARVIS-2D-STM-JPG/"
            + str(jid)
            + "_neg.jpg"
        )
        url = (
            "https://www.ctcms.nist.gov/"
            + "~knc6/static/JARVIS-DFT-STM/Images/JARVIS-2D-STM-JPG/"
        )
        if os.path.exists(stm_pos_file):
            stm_pos = url + str(jid) + "_pos.jpg"
        line += "<main_stm_pos>'" + str(stm_pos) + "'</main_stm_pos>"
        # f.write(str(line))
        if os.path.exists(stm_neg_file):
            stm_neg = url + str(jid) + "_neg.jpg"
        line += "<main_stm_neg>'" + str(stm_neg) + "'</main_stm_neg>"
        # f.write(str(line))
        line += "</basic_info>\n"
        # print (line)
        with open(filename, "w") as f:
            f.write(line)
        # f.write("</basic_info>\n")
        # f.close()
        time.sleep(wait_time)
        # line = (
        #    "<main_stm_neg>" + str(self.main_stm_neg()) + "</main_stm_neg>"
        # )
        # f.write(str(line))
        # line = (
        #    "<main_stm_pos>" + str(self.main_stm_pos()) + "</main_stm_pos>"
        # )
        # f.write(str(line))
        """
            line = "<stm_image>" + image_to_string() + "</stm_image>"+ "\n"
            f.write(str(line))
            line ='</basic_info>'
            f.write(str(line))
            """


"""
folder = '/rk2/knc6/JARVIS-DFT/2DRar-bulk/mp-17173_PBEBO'
filename='/users/knc6/Software/Devs/jarvis/jarvis/db/p.xml'
VaspToApiXmlSchema(folder=folder).write_xml(filename=filename)

folder = "/rk2/knc6/JARVIS-DFT/Met-Rar/mp-10891_PBEBO"
filename = "/cluster/users/knc6/justback/MTI/JVASP-16366.xml"
VaspToApiXmlSchema(folder=folder).write_xml(filename=filename)

folder = "/rk2/knc6/JARVIS-DFT/2D-1L/POSCAR-mp-2815-1L.vasp_PBEBO"
filename = "JVASP-664.xml"
VaspToApiXmlSchema(folder=folder).write_xml(filename=filename)


folder = "/rk2/knc6/JARVIS-DFT/Elements-bulkk/mp-149_bulk_PBEBO"
filename = "JVASP-1002.xml"
VaspToApiXmlSchema(folder=folder).write_xml(filename=filename)


if __name__ == "__main__":
    folder = "/rk2/knc6/JARVIS-DFT/TE-bulk/mp-541837_bulk_PBEBO"
    filename = "JVASP-1067.xml"
    VaspToApiXmlSchema(folder=folder).write_xml(filename=filename)

    folder = "/rk2/knc6/JARVIS-DFT/2DSF-bulk/mp-764365_PBEBO"
    filename = "JVASP-30266.xml"
    VaspToApiXmlSchema(folder=folder).write_xml(filename=filename)

    folder = "/rk2/knc6/JARVIS-DFT/Elements-bulkk/mp-149_bulk_PBEBO"
    filename = "JVASP-1002.xml"
    VaspToApiXmlSchema(folder=folder).write_xml(filename=filename)

    folder = "/rk2/knc6/JARVIS-DFT/2D-1L/POSCAR-mp-2815-1L.vasp_PBEBO"
    filename = "JVASP-664.xml"
    VaspToApiXmlSchema(folder=folder).write_xml(filename=filename)

    # directories = ["/rk2/knc6/JARVIS-DFT/Elements-bulkk/mp-149_bulk_PBEBO"]
    # filenames = ["JVASP-1002.xml"]
"""



"""Provide LAMMPS module .mod files."""

import os
import shutil


class GenericInputs(object):
    """Provide generic LAMMPS inputs such as relxation tolerance etc."""

    def __init__(self, input_name=""):
        """Intialize with filename."""
        self.input_name = input_name

    def elastic_general(self, path="."):
        """Provide filename for elastic constant calculations."""
        tmp_file = str(os.path.join(os.path.dirname(__file__), "inelast.mod"))
        shutil.copy2(tmp_file, path)

    def relax(self, path="."):
        """Provide filename for relaxation calculations."""
        tmp_file = str(os.path.join(os.path.dirname(__file__), "relax.mod"))
        shutil.copy2(tmp_file, path)

    def run0(self, path="."):
        """Provide filename for one-step calculations."""
        tmp_file = str(os.path.join(os.path.dirname(__file__), "run0.mod"))
        shutil.copy2(tmp_file, path)



from jarvis.analysis.solarefficiency.solar import SolarEfficiency
from jarvis.io.vasp.outputs import Vasprun
import os
import tarfile
import matplotlib.pyplot as plt
plt.switch_backend('agg')

example_fold_tgz = os.path.join(
    os.path.dirname(__file__),
    "..",
    "..",
    "..",
    "..",
    "examples",
    "vasp",
    "SiOptb88.tgz",
)


example_fold = os.path.join(
    os.path.dirname(__file__),
    "..",
    "..",
    "..",
    "..",
    "examples",
    "vasp",
    "SiOptb88",
)

if not os.path.isdir(example_fold):
    tar = tarfile.open(example_fold_tgz)
    tar.extractall(example_fold)
    tar.close()




vrun_path = os.path.join(
    os.path.dirname(__file__),
    "..",
    "..",
    "..",
    "..",
    "examples",
    "vasp",
    "SiOptb88",
    "SiOptb88",
    "MAIN-MBJ-bulk@mp_149",
    "vasprun.xml",
)


def test_solar():
    v = Vasprun(vrun_path)
    dirgap = v.get_dir_gap
    indirgap = v.get_indir_gap[0]
    en, abz = v.avg_absorption_coefficient
    abz = abz * 100
    eff_slme = SolarEfficiency().slme(
        en, abz, indirgap, indirgap, plot_current_voltage=True
    )
    # print("SLME", 100 * eff)
    eff_sq = SolarEfficiency().calculate_SQ(indirgap,  plot_current_voltage=True)
    # print("SQ", 100 * eff)
    assert (round(100 * eff_slme, 2), round(100 * eff_sq, 2)) == (33.23, 32.93)
    cmd='rm sq.png slme.png'
    os.system(cmd)

# test_solar()
